from credo.errors import NoConfigFile, BadConfigFile, BadConfiguration, ProgrammerError, CredoError, RepoError
from credo.asker import ask_for_choice, ask_for_choice_or_new, ask_for_ssh_key_folders
from credo.structure.credential_path import CredentialPath
from credo.crypto import Crypto
from credo import explorer

import logging
import json
import os

log = logging.getLogger("credo.overview")

class Unspecified(object):
    """Telling the difference between None and just not specified"""

class ConfigFileProperty(object):
    """A property that complains if it hasn't been set yet because setup must be called on Credo"""
    def __init__(self, name):
        self.name = "_{0}".format(name)

    def __get__(self, obj, type=None):
        if not hasattr(obj, self.name):
            raise ProgrammerError("Credo object needs to have setup() called before being used")
        return getattr(obj, self.name)

    def __set__(self, obj, val):
        setattr(obj, self.name, val)

class Credo(object):
    """Incredible credo knows all"""

    ########################
    ###   USAGE
    ########################

    @property
    def chosen(self):
        """Return our chosen creds"""
        if not hasattr(self, "_chosen"):
            self._chosen = self.make_chosen(rotate=True)
        return self._chosen

    def setup(self, config_file=Unspecified, root_dir=Unspecified, ssh_key_folders=Unspecified, providers=Unspecified, **kwargs):
        """Setup the credo!"""
        config_file, root_dir = self.find_config_file(config_file)

        if config_file:
            if not os.path.exists(config_file):
                raise NoConfigFile("Specified location is empty", location=config_file)
            if not os.access(config_file, os.R_OK):
                raise BadConfigFile("Config file isn't readable", location=config_file)

            self.read_from_config(config_file)

        # Override the root dir and ssh_key_folders if supplied
        self.config_file_location = config_file
        for key, val in (("root_dir", root_dir), ("ssh_key_folders", ssh_key_folders), ("providers", providers)):
            if val is not Unspecified:
                setattr(self, key, val)
            elif not hasattr(self, key):
                setattr(self, key, None)

        try:
            if self.root_dir and not os.path.exists(self.root_dir):
                os.makedirs(self.root_dir)
        except OSError as error:
            raise BadConfiguration("root_dir didn't exist and couldn't be made", root_dir=root_dir, error=error)

        self.set_options(**kwargs)
        self.validate_options()

    ########################
    ###   ATTRIBUTES SET BY OPTIONS
    ########################

    root_dir = ConfigFileProperty("root_dir")
    providers = ConfigFileProperty("providers")
    ssh_key_folders = ConfigFileProperty("ssh_key_folders")
    options_from_config = ["root_dir", "ssh_key_folders", "half_life", "providers"]

    def validate_options(self):
        """Make sure our options make sense"""
        errors = []
        if not self.root_dir:
            errors.append(BadConfiguration("Couldn't work out the root directory for your credentials..."))

        if self.ssh_key_folders:
            if not isinstance(self.ssh_key_folders, list):
                errors.append(BadConfiguration("ssh_key_folders is not a list", value_type=type(self.ssh_key_folders)))
            else:
                for folder in self.ssh_key_folders:
                    if not os.path.exists(folder):
                        errors.append(BadConfiguration("Given an ssh_key_folder that doesn't exist", folder=folder))

        if errors:
            raise BadConfiguration(errors=errors)

    ########################
    ###   CRYPTO
    ########################

    @property
    def crypto(self):
        """Memoize a crypto object"""
        if not getattr(self, "_crypto", None):
            self._crypto = self.make_crypto(self.ssh_key_folders)
        return self._crypto

    def make_crypto(self, ssh_key_folders=None):
        """Make the crypto object"""
        if not ssh_key_folders:
            home_ssh = os.path.expanduser("~/.ssh")
            if os.path.exists(home_ssh) and os.access(home_ssh, os.R_OK):
                ssh_key_folders = [home_ssh]
            else:
                ssh_key_folders = []

        crypto = Crypto()
        for folder in ssh_key_folders:
            crypto.find_keys(folder)

        # Keep asking for folders until we can sign
        while not crypto.has_private_keys():
            more_ssh_key_folders = ask_for_ssh_key_folders(already_have=crypto.ssh_key_folders)
            for folder in set(ssh_key_folders + crypto.ssh_key_folders + more_ssh_key_folders):
                crypto.find_keys(folder)

        return crypto

    ########################
    ###   CHOSEN CREDENTIALS
    ########################

    def make_chosen(self, rotate=True, invalidate_creds=False, half_life=None):
        """Make the chosen credentials from our repository"""
        structure, chains = self.find_credentials(asker=ask_for_choice)
        chosen = list(self.credentials_from(structure, chains, complain_if_missing=True))[0]
        if chosen.requires_encryption:
            chosen.credential_path.repository.pub_key_syncer.sync()

        if invalidate_creds:
            chosen.invalidate_creds()

        self.set_options(repo=chosen.credential_path.repository.name, account=chosen.credential_path.account.name, user=chosen.credential_path.user.name)

        if rotate:
            if chosen.keys.needs_rotation():
                changed = chosen.credential_path.repository.synchronize(override=True)
                if changed:
                    chosen.load()
                    chosen.credential_path.repository.pub_key_syncer.sync()

            chosen.save(half_life=half_life)
        return chosen

    def find_credentials(self, asker=None, missing_is_bad=False, want_new=False, no_mask=False):
        """
        Traverse our directory structure, asking as necessary

        and return the credentials object we find
        """
        directory_structure, shortened = explorer.find_repo_structure(self.root_dir, levels=3)
        if no_mask:
            mask = shortened
        else:
            mask = explorer.filtered(shortened, [self.repo, self.account, self.user], required_files=["credentials.json"])

        forced_vals = []
        if want_new:
            forced_vals = [self.repo, self.account, self.user]

        if asker:
            explorer.narrow(mask, ["Repository", "Account", "User"], asker, want_new=want_new, forced_vals=forced_vals)

        return directory_structure, explorer.flatten(directory_structure, mask, want_new=want_new)

    def credentials_from(self, directory_structure, chains, complain_if_missing=False, typ="amazon"):
        """Yield the credentials from the [location, <repo>, <account>, <user>] chains that are provided"""
        if not chains and complain_if_missing:
            raise CredoError("Didn't find any credentials!")

        for chain in chains:
            location, repo, account, user = chain
            credentials_location = os.path.join(location, "credentials.json")

            if not os.path.exists(credentials_location) and complain_if_missing:
                raise CredoError("Trying to find credentials that don't exist!", repo=repo, account=account, user=user)

            credential_path = CredentialPath(self.crypto)
            credential_path.fill_out(directory_structure, repo, account, user, typ=typ)
            credentials = credential_path.credentials
            credentials.load()
            yield credentials

    def find_one_repository(self, want_new=True):
        """Find one repository and return it's name and location"""
        _, shortened = explorer.find_repo_structure(self.root_dir, levels=1)
        mask = explorer.filtered(shortened, [self.repo])
        asker = ask_for_choice_or_new
        if not want_new:
            asker = ask_for_choice
        explorer.narrow(mask, ["Repository"], asker, want_new=want_new, forced_vals=[self.repo])
        if not mask:
            raise RepoError("Couldn't find a repository to work with.... try importing some keys....")

        repo_name = mask.keys()[0]
        location = os.path.join(self.root_dir, repo_name)
        return repo_name, location

    def find_credential_path_part(self, all_accounts=False, all_users=False, find_user=False, typ=None):
        """Find a repository, account or user"""
        directory_structure, shortened = explorer.find_repo_structure(self.root_dir, levels=3)
        mask = explorer.filtered(shortened, [self.repo, self.account, self.user])
        asker = ask_for_choice

        user_title = explorer.Stop if all_users else "User"
        account_title = explorer.Stop if all_accounts else "Account"
        want_any_after = 1
        if find_user:
            want_any_after = None
        explorer.narrow(mask, ["Repository", account_title, user_title], asker, want_any_after=want_any_after)
        if not mask:
            raise CredoError("Couldn't find anything to work with")

        repo = mask.keys()[0]
        user = None
        account = None
        if mask[repo]:
            account = mask[repo].keys()[0]
            if mask[repo][account]:
                user = mask[repo][account].keys()[0]

        credential_path = CredentialPath(self.crypto)
        credential_path.fill_out(directory_structure, repo, account, user, typ=typ)

        if credential_path.user:
            return credential_path.user
        elif credential_path.account:
            return credential_path.account
        else:
            return credential_path.repository

    ########################
    ###   CONFIGURATION
    ########################

    def set_options(self, **kwargs):
        """Set specific options"""
        repo = kwargs.get("repo")
        user = kwargs.get("user")
        account = kwargs.get("account")

        if account and "@" in account:
            if account.count("@") > 1:
                raise BadConfiguration("Account may only have one @ in it", account=account)
            account, repo = account.split("@")

        if user and "@" in user:
            if user.count("@") > 2:
                raise BadConfiguration("User may only have a max of two @ in it", user=user)
            elif user.count("@") == 2:
                user, account, repo = user.split("@")
            else:
                user, account = user.split("@")

        for attribute, val in (("user", user), ("account", account), ("repo", repo)):
            if getattr(self, attribute, None) is None:
                if val:
                    val = val.strip()
                setattr(self, attribute, val)

    def find_config_file(self, config_file=Unspecified):
        """Find a config file, use the one given if specified"""
        if config_file in (None, Unspecified):
            credo_home = os.path.expanduser("~/.credo")
            config_file = os.path.join(credo_home, "config.json")

        if os.path.exists(config_file):
            try:
                current = json.load(open(config_file))
            except (TypeError, ValueError) as error:
                raise CredoError("Couldn't parse the config file", location=config_file, error=error)
        else:
            current = {}

        if "root_dir" not in current:
            root_dir = os.path.join(os.path.dirname(config_file), "repos")
            current["root_dir"] = root_dir
            if not os.path.exists(os.path.dirname(config_file)):
                os.makedirs(os.path.dirname(config_file))
            json.dump(current, open(config_file, "w"))
        else:
            root_dir = current["root_dir"]

        return config_file, root_dir

    def read_from_config(self, config_file):
        """Call find_options using options from the config file"""
        options = json.load(open(config_file))
        for option in self.options_from_config:
            if option in options:
                setattr(self, option, options[option])

    def write_config(self):
        """Write the configuration"""
        cfg = dict((option, getattr(self, option, None)) for option in self.options_from_config)
        json.dump(cfg, open(self.config_file_location, "w"))

    ########################
    ###   SAML
    ########################

    def register_saml_provider(self, provider):
        """Register a saml provider"""
        if getattr(self, 'providers', None):
            self.providers.append(provider)
        else:
            self.providers = [provider]
        self.write_config()

    def remove_saml_provider(self, provider):
        """Remove a saml provider"""
        providers = getattr(self, 'providers', [])
        self.providers = [p for p in providers if p != provider]
        self.write_config()




from delfick_error import DelfickError, ProgrammerError

# Explicitly make ProgrammerError part of this context
ProgrammerError = ProgrammerError

class CredoError(DelfickError):
    """Anything inheriting this is caught and nicely displayed by executor:main"""

class NoExecCommand(CredoError):
    desc = "No exec command to execute"

class NoConfigFile(CredoError):
    desc = "No config file"

class BadConfigFile(CredoError):
    desc = "Bad config file"

class NotEnoughInfo(CredoError):
    desc = "Need more information"

class BadConfig(CredoError):
    desc = "Bad Config"

class BadCredentialFile(CredoError):
    desc = "Bad credentials file"

class BadSSHKey(CredoError):
    desc = "Bad ssh key"

class BadCypherText(CredoError):
    desc = "Couldn't decrypt text"

class BadConfiguration(CredoError):
    desc = "Bad configuration"

class NoCredentialsFound(CredoError):
    desc = "Couldn't find any credentials, please do an import"

class BadFolder(CredoError):
    desc = "Something wrong with a folder"

class CantEncrypt(CredoError):
    desc = "Can't do encryption"

class CantSign(CredoError):
    desc = "Can't sign data"

class BadPlainText(CredoError):
    desc = "Can't encrypt value"

class PasswordRequired(CredoError):
    desc = "Need a password"

class BadPrivateKey(BadSSHKey):
    desc = "Bad private ssh key"

class BadPublicKey(BadSSHKey):
    desc = "Bad public ssh key"

class NoSuchFingerPrint(CredoError):
    desc = "Haven't registered a key with specified fingerprint"

class CantFindPrivateKey(CredoError):
    desc = "Can't find any private key"

class NoValueEntered(CredoError):
    desc = "Couldn't get a value from the user"

class BadCredentialSource(CredoError):
    desc = "Bad source of credentials"

class BadCredential(CredoError):
    desc = "Invalid Credentials"

class UserQuit(CredoError):
    desc = "User quit the program"

class RepoError(CredoError):
    desc = "Something wrong with a repository"

class GitError(RepoError):
    desc = "Problem with git"

class InvalidData(CredoError):
    desc = "Invalid data"

class BadKeyFile(CredoError):
    desc = "Something wrong with a file containing keys"

class BadSamlProvider(CredoError):
    desc = "Something wrong with a saml provider"

class SamlNotAuthorized(CredoError):
    desc = "Saml said no"




from credo.cred_types.environment import EnvironmentMixin
from credo.helper import SignedValueFile

import logging
import os

log = logging.getLogger("credo.structure.account")

class Account(EnvironmentMixin):
    def __init__(self, name, location, credential_path):
        self.name = name
        self.location = location
        self.credential_path = credential_path

    @property
    def crypto(self):
        """Proxy credential_path"""
        return self.credential_path.crypto

    @property
    def repo_name(self):
        """Proxy credential_path"""
        return self.credential_path.repository.name

    @property
    def account_info_location(self):
        """Location of where our account id is"""
        return os.path.join(self.location, "account_id")

    @property
    def path(self):
        """Return the repo and account this represents"""
        return "repo={0}|account={1}".format(self.repo_name, self.name)

    @property
    def parent_path_part(self):
        """Return our repository"""
        return self.credential_path.repository

    def extra_env(self):
        """Define default env stuff"""
        return [("CREDO_CURRENT_ACCOUNT", self.name)]

    def account_id(self, suggestion=None, iam_pair=None):
        """
        Get us the account id for this account
        use suggestion and iam_pair to assist the user in choosing a value if there is none
        """
        if hasattr(self, "_account_id"):
            return self._account_id
        else:
            def suggestions():
                """Get us suggestions from suggestion and iam_pair"""
                result = []
                if suggestion:
                    result.append(suggestion)
                if iam_pair:
                    try:
                        result.append(iam_pair.ask_amazon_for_account())
                    except:
                        pass
                return result

            info_location = self.account_info_location
            signed_value_file = SignedValueFile(info_location, self.crypto, dict(repo=self.repo_name, account=self.name))
            question = "How do you want to enter the account id?\trepo={0}\taccount={1}".format(self.repo_name, self.name)
            account_id, created = signed_value_file.retrieve("Account ID", question, suggestions)

            if created:
                self.credential_path.add_change("Writing account id {0}".format(account_id), [info_location], repo=self.repo_name, account=self.name)

            self._account_id = account_id

        return self._account_id




from credo.asker import ask_for_choice, ask_for_choice_or_new
from credo.cred_types.environment import EnvironmentMixin
from credo.versioning import has_git_abilities
from credo.errors import UserQuit, RepoError
from credo.versioning import determine_driver
from credo.pub_keys import PubKeySyncer

import logging
import os

log = logging.getLogger("credo.structure.repository")

def synchronize(repo_name, location, crypto):
    """Synchronise this repository"""
    repository = Repository(repo_name, location, crypto)
    if repository.versioned:
        repository.synchronize()

def configure(repo_name, location, crypto, new_remote=None, version_with=None):
    """Configure a repository"""
    repository = Repository(repo_name, location, crypto)
    if repository.versioned:
        remote = "<none set>"
        if repository.remote:
            remote = repository.remote

        remove_choice = "Stop versioning this directory"
        change_choice = "Change the remote of this directory"
        carryon_choice = "Keep the remote as is"

        if version_with == "nothing":
            choice = remove_choice
        elif new_remote is None and version_with is None:
            choice = ask_for_choice(
                [ "This repo is already versioned\tremote={0}\tlocation={1}".format(remote, repository.location)
                , "What do you want to do?"
                ]
                , [remove_choice, change_choice, carryon_choice]
                )
        else:
            choice = change_choice

        if choice == remove_choice:
            repository.deleteme()
        elif choice == change_choice:
            repository.change_remote(new_remote, remote_type=version_with)
    else:
        change_choice = "Add versioning to this repository"
        carryon_choice = "Keep the repository unversioned"

        if version_with == "nothing":
            log.info("Repository is already unversioned")
        elif version_with is None:
            choice = ask_for_choice(
                [ "This repo is not versioned yet ({0})".format(repository.location)
                , "What do you want to do?"
                ]
                , [carryon_choice, change_choice]
                )
        else:
            choice = change_choice

        if choice == change_choice:
            repository.change_remote(new_remote, remote_type=version_with)

class Repository(EnvironmentMixin):
    """Understands how to version a directory"""
    def __init__(self, name, location, crypto):
        self.name = name
        self.crypto = crypto
        self.location = location

        self.driver = determine_driver(location)

        root_dir = os.path.join(os.path.dirname(self.location), "..")
        self.pub_key_syncer = PubKeySyncer(root_dir, self)

    def synchronize(self, override=False):
        """Ask the driver to synchronize the folder"""
        self.driver.synchronize(override=override)

    @property
    def path(self):
        """Return the repo this represents"""
        return "repo={0}".format(self.name)

    @property
    def path_parts(self):
        """Return the objects in the path to here"""
        return ()

    def extra_env(self):
        """Define default env stuff"""
        return [("CREDO_CURRENT_REPO", self.name)]

    @property
    def versioned(self):
        """Say whether the driver is versioned or not"""
        return self.driver.versioned

    @property
    def remote(self):
        """Give back the remote the driver has"""
        return self.driver.remote

    def deleteme(self):
        """Tell the driver to delete himself"""
        self.driver.deleteme()

    def change_remote(self, new_remote=None, remote_type=None):
        """
        Determine what we want to change the remote to

        If passed in new_remote is None, we ask for a remote
        Only support git, so complain if remote_type is not "git", and set it to git if it is None
        """
        message = "What do you want as the new remote?"
        if self.driver.versioned:
            message = "{0} (currently {1}".format(message, self.remote)

            no_remote_choice = "Keep versioned, but remove our registered remote"
        else:
            no_remote_choice = "Setup versioning, but don't add a remote"

        quit_choice = "Quit"

        if new_remote is False:
            choice = no_remote_choice
        elif new_remote is not None:
            choice = new_remote
        else:
            choice = ask_for_choice_or_new(message, [quit_choice, no_remote_choice])

        if choice == quit_choice:
            raise UserQuit()
        elif choice == no_remote_choice:
            choice = None

        if remote_type is None:
            remote_type = "git"
            if not has_git_abilities():
                raise RepoError("Sorry, no pygit2, can't do git things")
        elif remote_type != "git":
            raise RepoError("Unsupported versioning type", type=remote_type)

        if not self.driver.versioned:
            self.driver = determine_driver(self.location, version_type=remote_type)
            self.driver.initialise(choice)
        else:
            self.driver.change_remote(choice)

    def add_change(self, message, changed_files, **info):
        """Ask the driver to add the changed files and commit with the provided message"""
        message_suffix = ", ".join("{0}={1}".format(key, val) for key, val in info.items())
        if message_suffix:
            message = "{0} ({1})".format(message, message_suffix)

        changes = []
        for filename in changed_files:
            if filename.startswith("/"):
                changes.append(os.path.relpath(filename, start=self.location))
            else:
                changes.append(filename)

        self.driver.add_change(message, changes)




from credo.helper import KeysFile

class Keys(object):
    """Collection of environment variables"""
    def __init__(self, location, credential_path):
        self.location = location
        self.credential_path = credential_path
        self.crypto = credential_path.crypto
        self._changed = False

    def add(self, key, value):
        """Add a key"""
        raise NotImplementedError()

    def unchanged(self):
        """Reset changed on everything"""
        raise NotImplementedError()

    @property
    def encrypted_values(self):
        """Return our values as a dictionary with encrypted values"""
        raise NotImplementedError()

    def make_keys(self, contents):
        """Get us our keys from the contents of the file"""
        raise NotImplementedError()

    def shell_exports(self):
        """Return list of (key, val) exports we want to have in the shell"""
        env_list = []
        if hasattr(self, "extra_env"):
            env_list.extend(self.extra_env())
        if getattr(self, "parent_path_part", None):
            env_list.extend(self.parent_path_part.shell_exports())
        if hasattr(self, "exports"):
            env_list.extend(self.exports())
        return env_list

    def as_string(self):
        """Return information about keys as a string"""
        raise NotImplementedError()

    @property
    def default_keys_type(self):
        """Return the default key type (i.e. list, dict)"""
        raise NotImplementedError()

    @property
    def default_keys_type_name(self):
        """Return the default key type name (i.e. amazon, environment)"""
        raise NotImplementedError()

    def load(self):
        """Just return the contents"""
        return self.contents

    def save(self, force=False):
        """Save our credentials to file"""
        self.contents.save(self.location, self)

    def __iter__(self):
        """Iterate through the environment variables"""
        return iter(self.keys.items())

    def __len__(self):
        """Return how many environment variables we have"""
        return len(self.keys)

    @property
    def changed(self):
        """Say whether there has been any changes"""
        return self._changed

    @property
    def contents(self):
        """Get us the contents"""
        if not hasattr(self, "_contents"):
            self._contents = KeysFile(default_keys_type=self.default_keys_type, default_keys_type_name=self.default_keys_type_name)
            self._contents.load(self.location)
        return self._contents

    @property
    def keys(self):
        """Get us some keys"""
        if not hasattr(self, "_keys"):
            self._keys = self.make_keys(self.contents)
        return self._keys




from credo.errors import BadSamlProvider, CredoError, SamlNotAuthorized
from credo.asker import ask_user_for_half_life
from credo.cred_types.saml import SamlRole

from boto.iam.connection import IAMConnection
from boto.sts.connection import STSConnection
from boto.sts.credentials import AssumedRole
from datetime import datetime, timedelta
import xml.etree.ElementTree as ET
from six.moves import http_client
from textwrap import dedent
import requests
import xml.sax
import logging
import base64
import time
import uuid
import boto
import os

log = logging.getLogger("credo.amazon")

########################
###   BASE
########################

class IamBase(object):
    def __init__(self, create_epoch=None, half_life=None):
        self.setup(create_epoch=create_epoch, half_life=half_life)

    def setup(self, create_epoch=None, half_life=None):
        """Setup the pair"""
        self._half_life = half_life
        self._create_epoch = create_epoch
        self._changed = False
        self.deleted = False

    @property
    def changed(self):
        """Get us value of _changed"""
        return self._changed

    def unchanged(self):
        """Set _changed to False"""
        self._changed = False

    def mark_as_invalid(self):
        """Mark the key as invalid"""
        self.invalidated = True

    @property
    def works(self):
        """Says whether this key is valid enough to get iam informations"""
        if self.deleted:
            return True
        if getattr(self, "_works", None) is False:
            return self._works

        get_cached = hasattr(self, "_last_asked") and (time.time() - self._last_asked) < 6
        self._last_asked = time.time()
        self._get_info(get_cached=get_cached)

        return self._works

    def wait_till_works(self):
        """Wait till this iam pair works"""
        # Give amazon time to think about this
        start = time.time()
        while time.time() - start < 20:
            self._get_info(quiet=True)
            if self._works:
                break
            time.sleep(2)

    def ask_amazon_for_account(self):
        """Get the account id for this key"""
        self._get_info(get_cached=True)
        return getattr(self, "account_id", None)

    def ask_amazon_for_account_aliases(self):
        """Get the account aliases for this key"""
        self._get_info(get_cached=True)
        return getattr(self, "account_aliases", None)

    def ask_amazon_for_username(self):
        """Get the username for this key"""
        self._get_info(get_cached=True)
        return getattr(self, "username", None)

    @property
    def create_epoch(self):
        """Use our create_epoch or ask amazon for it"""
        if self._create_epoch:
            return self._create_epoch
        elif self.works:
            self._create_epoch = self.ask_amazon_for_create_epoch()
            self._changed = True
            return self._create_epoch
        else:
            return 0

    @property
    def half_life(self):
        """Use our half_life or ask user for one"""
        if not self._half_life:
            self._half_life = ask_user_for_half_life(self.aws_access_key_id)
            self._changed = True
        return self._half_life

    def set_half_life(self, half_life):
        """Record a new half_life"""
        if half_life != self._half_life:
            self._half_life = half_life
            self._changed = True

    def is_root_credentials(self):
        """
        Return whether these credentials are possibly root credentials
        I.e. Amazon say they don't exist and it has the same name as an account alias
        """
        username = self.ask_amazon_for_username()
        try:
            self.connection.get_all_access_keys(username)
        except boto.exception.BotoServerError as error:
            if error.status == 404 and error.code == "NoSuchEntity":
                if username in self.ask_amazon_for_account_aliases():
                    return True
            else:
                raise
        return False

    def amazon_date_to_epoch(self, create_date):
        """Convert create_date from amazon into a create_epoch"""
        dt = boto.utils.parse_ts(create_date)
        return (dt - datetime(1970, 1, 1)).total_seconds()

    def expired(self):
        """Say whether the age of this key is past twice it's half life or marked as invalid"""
        return getattr(self, "invalidated", False) or self.half_life > 0 and self.age > self.half_life * 2

    def past_half_life(self):
        """Say whether the age of this key is past it's half life"""
        return self.half_life > 0 and self.age > self.half_life

    @property
    def age(self):
        """Age is time since it's create_epoch"""
        return time.time() - self.create_epoch

########################
###   USER CREDS
########################

class IamPair(IamBase):
    def __init__(self, aws_access_key_id, aws_secret_access_key, aws_security_token=None, create_epoch=None, half_life=None):
        self.aws_access_key_id = aws_access_key_id
        self.aws_security_token = aws_security_token
        self.aws_secret_access_key = aws_secret_access_key
        super(IamPair, self).__init__(create_epoch=create_epoch, half_life=half_life)

    @property
    def connection(self):
        """Get a connection for these keys"""
        if not getattr(self, "_connection", None):
            self._connection = IAMConnection(self.aws_access_key_id, self.aws_secret_access_key, security_token=self.aws_security_token)
        return self._connection

    def _get_info(self, get_cached=False, quiet=False):
        """
        Get user details from this key and set
        self._works
        self.username
        self.account_id
        """
        try:
            if getattr(self, "_got_user", None) is None or not get_cached:
                log.info("Asking amazon for account id and username\taccess_key=%s", self.aws_access_key_id)
                details = self.connection.get_user()["get_user_response"]["get_user_result"]["user"]
                aliases = self.connection.get_account_alias()["list_account_aliases_response"]["list_account_aliases_result"]["account_aliases"]
                self._invalid = False
                self._works = True
                self._got_user = True
                self.username = details["user_name"]

                # arn is arn:aws:iam::<account_id>:<other>
                self.account_id = details["arn"].split(":")[4]
                self.account_aliases = aliases
        except boto.exception.BotoServerError as error:
            self._works = False
            self._got_user = False
            self._connection = None
            if error.status == 403 and error.code in ("InvalidClientTokenId", "SignatureDoesNotMatch"):
                self._invalid = True
                if not quiet:
                    log.info("Found invalid access key and secret key combination\taccess_key=%s\terror_code=%s\terror=%s", self.aws_access_key_id, error.code, error.message)
                return
            raise

    def find_other_access_keys(self):
        """Find all the access_keys for this user"""
        keys = self.connection.get_all_access_keys(self.ask_amazon_for_username())["list_access_keys_response"]["list_access_keys_result"]["access_key_metadata"]
        return [str(key["access_key_id"]) for key in keys]

    def ask_amazon_for_create_epoch(self):
        """Return our create_epoch"""
        if self.is_root_credentials():
            result = self.connection.get_response('ListAccessKeys', {}, list_marker='AccessKeyMetadata')
        else:
            username = self.ask_amazon_for_username()
            result = self.connection.get_all_access_keys(username)

        access_keys = result["list_access_keys_response"]["list_access_keys_result"]["access_key_metadata"]
        create_date = [key for key in access_keys if key["access_key_id"] == self.connection.aws_access_key_id][0]['create_date']
        return self.amazon_date_to_epoch(create_date)

    def create_new(self):
        """Create a new iam pair to use"""
        log.info("Creating a new key")
        response = self.connection.create_access_key(self.ask_amazon_for_username())["create_access_key_response"]["create_access_key_result"]["access_key"]
        log.info("Created %s", response["access_key_id"])
        iam_pair = IamPair(str(response["access_key_id"]), str(response["secret_access_key"]), create_epoch=self.amazon_date_to_epoch(response["create_date"]))
        iam_pair.wait_till_works()
        return iam_pair

    def delete(self):
        """Delete this key pair from amazon"""
        return self.delete_access_key(self.aws_access_key_id)

    def delete_access_key(self, access_key):
        log.info("Deleting a key\taccess_key_id=%s", access_key)
        if access_key == self.aws_access_key_id:
            self.deleted = True
            self._changed = True
        return self.connection.delete_access_key(access_key)

    @classmethod
    def from_environment(kls, create_epoch=None, half_life=None):
        """Get an IAMPair from our environment variables"""
        pair = kls(os.environ["AWS_ACCESS_KEY_ID"], os.environ["AWS_SECRET_ACCESS_KEY"], os.environ.get("AWS_SECURITY_TOKEN"), create_epoch, half_life)
        pair._connection = IAMConnection()
        return pair

    def synchronize_with(self, other):
        """Synchronise non key information with another key"""
        self.set_half_life(other._half_life)

########################
###   SAML CREDS
########################

class FixedSTSConnection(STSConnection):
    """
    assume_role_with_saml seems broken because it puts the assertion in the query parameters.
    awscli puts it in the POST data and isn't broken, so lets do that here as well.
    """
    def assume_role_with_saml(self, role_arn, principal_arn, saml_assertion, policy=None, duration_seconds=None):
        data = {
            'RoleArn': role_arn,
            'PrincipalArn': principal_arn,
            'SAMLAssertion': saml_assertion,
        }
        if policy is not None:
            data['Policy'] = policy
        if duration_seconds is not None:
            data['DurationSeconds'] = duration_seconds

        params = {"Action": "AssumeRoleWithSAML", "Version": self.APIVersion}
        response = requests.post("https://{0}".format(self.host), headers={"User-Agent": boto.UserAgent}, params=params, data=data)

        if response.status_code == 200:
            obj = AssumedRole(self)
            h = boto.handler.XmlHandler(obj, self)
            xml.sax.parseString(response.content, h)
            return obj
        else:
            boto.log.error('%s %s' % (response.status_code, response.reason))
            boto.log.error('%s' % response.content)
            raise self.ResponseError(response.status_code, response.reason, response.content)

class IamSaml(IamBase):
    def __init__(self, provider, username, password, connection=None, create_epoch=None, half_life=None):
        self.provider = provider
        self.basic_auth = base64.b64encode("{0}:{1}".format(username, password).encode('utf-8'))
        self.username = username
        self._connection = connection
        super(IamSaml, self).__init__(create_epoch=create_epoch, half_life=half_life)

    @property
    def connection(self):
        """Get a connection"""
        if not getattr(self, "_connection", None):
            self._connection = FixedSTSConnection(anon=True)
        return self._connection

    @property
    def arns(self):
        if not getattr(self, "_arns", None):
            self._get_info()
        return self._arns

    @property
    def assertion(self):
        if not getattr(self, "_assertion", None):
            self._get_info()
        return self._assertion

    def _get_info(self, get_cached=False, quiet=False):
        """
        Authenticate against the provided username and password

        Create some variables:
            idpid = your idp entity id
            rpid = a valid SP entityId that is configured for ECP
            acsurl,ascurlbinding = an AssertionConsumerService URL and binding
                this will match the values in the 'relying party' and associated metadata

        And then make a soap request with base64 encoded ldap username and password
        and get back the authentication string
        """
        idpid = "https://{0}/idp/shibboleth".format(self.provider)
        rpid = "urn:amazon:webservices"
        acsurl = "https://signin.aws.amazon.com/saml"
        acsurlbinding = "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"

        tree = None
        tree_body = None
        for _ in range(5):
            now = (datetime.utcnow() - timedelta(minutes=1)).strftime("%Y-%m-%dT%H:%M:%S")
            ident = uuid.uuid1().hex
            envelope = dedent("""
            <S:Envelope xmlns:S="http://schemas.xmlsoap.org/soap/envelope/" xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol">
                <S:Body>
                    <samlp:AuthnRequest
                        xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                        AssertionConsumerServiceURL="{acsurl}"
                        ID="_{ident}"
                        IssueInstant="{now}"
                        ProtocolBinding="{acsurlbinding}"
                        Version="2.0"
                    >
                        <saml:Issuer xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">{rpid}</saml:Issuer>
                        <samlp:NameIDPolicy AllowCreate="1"/>
                        <samlp:Scoping>
                            <samlp:IDPList>
                                <samlp:IDPEntry ProviderID="{idpid}"/>
                            </samlp:IDPList>
                        </samlp:Scoping>
                    </samlp:AuthnRequest>
                </S:Body>
            </S:Envelope>
            """).format(acsurl=acsurl, acsurlbinding=acsurlbinding, ident=ident, now=now, idpid=idpid, rpid=rpid)

            headers = {"Accept": "*/*", "Authorization": "Basic {0}".format(self.basic_auth.decode('utf-8')), "Content-Type": "application/x-www-form-urlencoded", "Content-Length": len(envelope)}
            connection = http_client.HTTPSConnection(self.provider, 443)
            connection.request("POST", "/idp/profile/SAML2/SOAP/ECP", envelope, headers)
            resp = connection.getresponse()

            if resp.status == 401:
                log.info("Not authorized, perhaps you entered the wrong password")
                self._works = False
                raise SamlNotAuthorized()
            elif resp.status != 200:
                log.info("Failed to authenticate, trying again in 5 seconds")
                time.sleep(5)
                self._works = False
            else:
                body = resp.read().decode('utf-8')
                tree = ET.fromstring(body)
                tree_body = tree.find("{http://schemas.xmlsoap.org/soap/envelope/}Body")
                if tree_body is None:
                    log.info("Failed to authenticate, saml provider didn't return any soap. Trying again in 5 seconds")
                    time.sleep(5)
                    self._works = False
                else:
                    self._works = True
                    break

        if tree is None:
            raise BadSamlProvider("Failed to authenticate", provider=self.provider, username=self.username)

        arns = tree.findall(".//*[@FriendlyName='Role']/*")
        self._works = True
        body_start = "<soap11:Body>"
        body_end = "</soap11:Body>"
        self._assertion = base64.b64encode(body[body.find(body_start)+len(body_start):body.find(body_end)].encode('utf-8'))
        self._arns = sorted([SamlRole(*arn.text.split(",")) for arn in arns], key = lambda a: a.role_arn)

    def exports(self, role):
        """Get exports for this account"""
        if role.role_arn not in [r.role_arn for r in self.arns]:
            raise CredoError("Your user doesn't have specified account anymore"
                , username=self.keys.idp_username, provider=self.keys.provider, wanted=self.keys.role.role_arn
                )

        result = self.get_result(role)
        creds = result.credentials
        return [
              ("AWS_ACCESS_KEY_ID", creds.access_key)
            , ("AWS_SECRET_ACCESS_KEY", creds.secret_key)

            , ("AWS_SESSION_TOKEN", creds.session_token)
            , ("AWS_SECURITY_TOKEN", creds.session_token)
            ]

    def get_result(self, role):
        """Get back the sts assume result"""
        return self.connection.assume_role_with_saml(role.role_arn, role.principal_arn, self.assertion, duration_seconds=3600)

    def synchronize_with(self, other):
        """Nothing to synchronize"""




from credo.structure.credentials import Credentials, SamlCredentials
from credo.structure.repository import Repository
from credo.structure.account import Account
from credo.structure.user import User

import logging
import json
import os

log = logging.getLogger("credo.structure.credential_path")

class CredentialPath(object):
    """Knows about everything leading up to a credentials"""
    user = None
    account = None
    repository = None
    credentials = None

    def __init__(self, crypto):
        self.crypto = crypto

    def fill_out(self, directory_structure, repo, account, user, typ="amazon"):
        """Make the things leading up to the credentials"""
        self.repository = Repository(repo, directory_structure[repo]['/location/'], self.crypto)
        if account:
            self.account = Account(account, directory_structure[repo][account]['/location/'], self)
            if user:
                self.user = User(user, directory_structure[repo][account][user]['/location/'], self)
                credential_location = os.path.join(self.user.location, "credentials.json")
                self.credentials = self.make_credentials(credential_location, self, typ)

    def make_credentials(self, location, cred_path, typ="amazon"):
        """Make a credentials object from this location"""
        if os.path.exists(location):
            with open(location) as f:
                contents = None
                try:
                    contents = json.load(f)
                except (ValueError, TypeError) as error:
                    log.warning("Failed to find type of file\terror=%s\tlocation=%s", error, location)

                if contents and isinstance(contents, dict) and 'type' in contents:
                    typ = contents.get("type", typ)

        if typ == "saml":
            return SamlCredentials(location, self)
        else:
            return Credentials(location, self)

    def add_change(self, location, message, **info):
        """Register a change that was made"""
        self.repository.add_change(location, message, **info)




from credo.actions import (
      do_exec, do_showavailable, do_import
    , do_rotate, do_current, do_remote
    , do_synchronize, do_capture, do_env
    , do_unset, do_register_saml, do_serve
    , do_switch, do_output_extension, do_exports
    , do_print_shell_function, do_create_launch_daemon
    )
from credo.errors import CredoError, NoExecCommand
from credo.asker import secret_sources
from credo.overview import Credo
from credo import VERSION

from rainbow_logging_handler import RainbowLoggingHandler
import argparse
import logging
import sys
import os

log = logging.getLogger("executor")

def setup_logging(verbose=False, boto_debug=False):
    log = logging.getLogger("")
    handler = RainbowLoggingHandler(sys.stderr)
    handler._column_color['%(asctime)s'] = ('cyan', None, False)
    handler._column_color['%(levelname)-7s'] = ('green', None, False)
    handler._column_color['%(message)s'][logging.INFO] = ('blue', None, False)
    handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)-7s %(name)-15s %(message)s"))
    log.addHandler(handler)
    log.setLevel([logging.INFO, logging.DEBUG][verbose])

    if boto_debug:
        logging.getLogger("boto").setLevel(logging.DEBUG)
    else:
        logging.getLogger("boto").setLevel([logging.CRITICAL, logging.ERROR][verbose])

    logging.getLogger("requests").setLevel([logging.CRITICAL, logging.ERROR][verbose])

def sourceable_if(action_is):
    """Set a parser as sourceable if the action is as specified"""
    def wrapper(func):
        func.sourceable_if = action_is
        return func
    return wrapper

class CliParser(object):
    def split_argv(self, argv=None):
        """Split the args into (cred_opts, action, action_opts)"""
        if argv is None:
            argv = sys.argv[1:]

        for index, arg in enumerate(argv):
            if arg in self.actions:
                return argv[:index+1], arg, argv[index+1:]

        # Couldn't find a valid action :(
        return argv, None, []

    def parse_args(self, argv=None):
        """
        Get us (credo, kwargs, function)

        Where credo is an overview object of our credo collection

        kwargs is the extra arguments to call the function with

        And function is what we want to call with the kwargs
        The function should have the signature function(credo, **kwargs)
        """
        cred_args, action, action_args = self.split_argv(argv)
        if "--version" in cred_args:
            self.show_version_and_quit()

        credo = self.make_credo(cred_args, action)
        kwargs, function = self.actions[action](action, action_args)
        return credo, kwargs, function

    @property
    def actions(self):
        return {
              "help": self.parse_help
            , "exec": self.parse_exec
            , "show": self.parse_show
            , "serve": self.parse_serve
            , "switch": self.parse_switch
            , "remote": self.parse_remote
            , "import": self.parse_import
            , "rotate": self.parse_rotate
            , "version": self.parse_version
            , "sourceable": self.parse_sourceable
            , "register_saml": self.parse_register_saml
            , "output_extension": self.parse_output_extension
            , "print_shell_function": self.parse_print_shell_function
            , "create_launch_daemon": self.parse_create_launch_daemon

            , "env": self.parse_env
            , "capture": self.parse_env

            , "unset": self.parser_for_no_args("Unset credo environment variables", do_unset, sourceable=True)
            , "inject": self.parser_for_no_args("Print out export statements for your aws creds", do_exports, sourceable=True, extra_args=self.other_export_args)
            , "exports": self.parser_for_no_args("Print out export statements for your aws creds", do_exports, extra_args=self.other_export_args)
            , "current": self.parser_for_no_args("Show what user is currently in your environment", do_current)
            , "synchronize": self.parser_for_no_args("Synchronise with the remote for some repository", do_synchronize)
            }

    def cred_parser(self):
        """Parser for all the common credo options"""
        parser = argparse.ArgumentParser(description="Credo executor")

        parser.add_argument("action"
            , help = "What should credo do today?"
            , choices = self.actions
            )

        parser.add_argument("-a", "--account"
            , help = "The account to use (optionally use account@repo syntax)"
            )

        parser.add_argument("-u", "--user"
            , help = "The user to use (optionally use user@account@repo syntax, where account and repo are optional)"
            )

        parser.add_argument("-r", "--repo"
            , help = "The repo to use"
            )

        parser.add_argument("-v", "--verbose"
            , help = "Show debug log messages"
            , action = "store_true"
            )

        parser.add_argument("--version"
            , help = "Print version and quit"
            , dest = "show_version"
            , action = "store_true"
            )

        parser.add_argument("--boto-debug"
            , help = "Show debug log messages for boto"
            , action = "store_true"
            )

        def folder(value):
            location = os.path.expanduser(value)
            if not os.path.isfile(location):
                raise CredoError("Please specify a config file that exists", specified=location)
            return location

        parser.add_argument("--config"
            , help = "The location to the config file"
            , dest = "config_file"
            , type = folder
            )

        return parser

    def show_version_and_quit(self):
        """Show the version and quit"""
        print("Credo {0}".format(VERSION))
        sys.exit(0)

    def args_from_subparser(self, action, parser, argv):
        """Get us args from our parser as a dictionary and make sure our usage statement is nice"""
        cred_usage = self.cred_parser().format_usage().split(":", 1)[1].strip().split("\n")[0]
        subparser_usage = parser.format_usage()
        subparser_usage = subparser_usage[subparser_usage.index(parser.prog)+len(parser.prog):].strip()
        parser.usage = "{0} <|| {1} ||> {2}".format(cred_usage, action, subparser_usage)
        return vars(parser.parse_args(argv))

    def make_credo(self, cred_args, expected_action):
        """Make a Credo object that knows things"""
        cred_parser = self.cred_parser()
        cred_args = cred_parser.parse_args(cred_args)

        setup_logging(verbose=cred_args.verbose, boto_debug=cred_args.boto_debug)

        if cred_args.action != expected_action:
            raise CredoError("Well this is weird, I thought the action was different than it turned out to be", expected=expected_action, parsed=cred_args.action)

        credo = Credo()
        credo.setup(**vars(cred_args))
        return credo

    def parser_for_no_args(self, description, func, sourceable=False, extra_args=None):
        """Return a function that parses no arguments"""

        def parse_noargs(action, argv):
            """No args to parse"""
            parser = argparse.ArgumentParser(description=description)
            if sourceable:
                parser.add_argument("--no-sourcing"
                    , help = "Tell credo sourceable not to source this output"
                    , action = "store_true"
                    )
            if extra_args:
                parser = extra_args(action, argv, parser)
            args = self.args_from_subparser(action, parser, argv)
            return args, func
        parse_noargs.sourceable = sourceable

        return parse_noargs

    def other_export_args(self, action, argv, parser):
        parser.add_argument("--half-life"
            , help = "Choose a half life for your new key"
            , choices = ["hour", "day", "week"]
            )
        return parser

    def parse_help(self, action, argv):
        """Just prints help and quits"""
        # It's late, I'm tired....
        print("Try the --help option")
        sys.exit(1)

    def parse_version(self, action, argv):
        """Just show the version and quit"""
        self.show_version_and_quit()

    @sourceable_if("env")
    def parse_env(self, action, argv):
        """Capture/display environment variables"""
        parser = argparse.ArgumentParser(description="Capture environment variables")

        def env_spec(key):
            """Specification for an environment variable"""
            if "=" not in key:
                if key not in os.environ:
                    raise argparse.ArgumentTypeError("The specified environment variable {0} isn't in the current environment".format(key))
                else:
                    return (key, os.environ[key])
            else:
                k, v = key.split('=', 1)
                return (k.strip(), v.strip())

        if action == "capture":
            parser.add_argument("--env"
                , help = "Capture this environment variable"
                , type = env_spec
                , action = "append"
                )

            parser.add_argument("--remove-env"
                , help = "Remove this env from our capture"
                , action = "append"
                )

        chooser = parser.add_mutually_exclusive_group()

        chooser.add_argument("--all-accounts"
            , help = "Capture environment variables for our chosen repository"
            , action = "store_true"
            )

        chooser.add_argument("--all-users"
            , help = "Capture environment variables for our chosen account"
            , action = "store_true"
            )

        chooser.add_argument("--find-user"
            , help = "Find a particular user"
            , action = "store_true"
            )

        if action == "env":
            parser.add_argument("--no-sourcing"
                , help = "Make credo sourceable say this is not sourceable"
                , action = "store_true"
                )

        args = self.args_from_subparser(action, parser, argv)

        func = do_env
        if action == "capture":
            func = do_capture
        return args, func

    def parse_rotate(self, action, argv):
        """Rotate our credentials"""
        parser = argparse.ArgumentParser(description="Rotate the credentials")
        parser.add_argument("--force"
            , help = "Force expire everything"
            , action = "store_true"
            )
        parser.add_argument("--half-life"
            , help = "Choose a half life for your new key"
            , choices = ["hour", "day", "week"]
            )
        args = self.args_from_subparser(action, parser, argv)
        return args, do_rotate

    def parse_output_extension(self, action, argv):
        """Rotate our credentials"""
        parser = argparse.ArgumentParser(description="Output the chrome extension for the metadata server.")
        parser.add_argument("--output"
            , help = "The folder where the extension will be saved to."
            , required = True
            )
        args = self.args_from_subparser(action, parser, argv)
        return args, do_output_extension

    def parse_print_shell_function(self, action, argv):
        """Show our credo bash function"""
        parser = argparse.ArgumentParser(description="The current credo bash function")
        parser.add_argument("--virtualenv"
            , help = "The virtualenv where credo is installed."
            )
        args = self.args_from_subparser(action, parser, argv)
        return args, do_print_shell_function

    def parse_create_launch_daemon(self, action, argv):
        """Create the LaunchDaemon plist file."""
        parser = argparse.ArgumentParser(description="Create the LaunchDaemon plist file.")
        parser.add_argument("--virtualenv"
            , help = "The virtualenv where credo is installed."
            )
        args = self.args_from_subparser(action, parser, argv)
        return args, do_create_launch_daemon

    def parse_show(self, action, argv):
        """Parser for showing available credentials"""
        parser = argparse.ArgumentParser(description="Show you the credentials you have")
        parser.add_argument("--all"
            , help = "Force show all available"
            , action = "store_true"
            , dest = "force_show_all"
            )
        parser.add_argument("--no-collapse"
            , help = "Don't collapse output if we only find one"
            , action = "store_false"
            , dest = "collapse_if_one"
            )
        args = self.args_from_subparser(action, parser, argv)
        return args, do_showavailable

    def parse_import(self, action, argv):
        """Import doesn't have any arguments yet"""
        parser = argparse.ArgumentParser(description="Import amazon secrets")
        parser.add_argument("--source"
            , help = "Choose a particular source to get credentials from"
            , choices = secret_sources.keys()
            )
        parser.add_argument("--half-life"
            , help = "Choose a half life for your new key"
            , choices = ["hour", "day", "week"]
            )
        args = self.args_from_subparser(action, parser, argv)
        return args, do_import

    def parse_remote(self, action, argv):
        """Options for setting an external remote for syncing with"""
        parser = argparse.ArgumentParser(description="Set up a remote for a repository")

        parser.add_argument("--version-with"
            , help = "Use this to make this repository versioned"
            , choices = ["nothing", "git"]
            )

        remote = parser.add_mutually_exclusive_group()
        remote.add_argument("--no-new-remote"
            , help = "Don't set a remote"
            , dest = "remote"
            , const = False
            , action = "store_const"
            )
        remote.add_argument("--remote"
            , help = "Set a particular remote"
            )

        args = self.args_from_subparser(action, parser, argv)
        return args, do_remote

    def parse_exec(self, action, argv):
        """Exec passes on everything else also doesn't have arguments yet"""
        parser = argparse.ArgumentParser(description="Run the provided command using a sub shell with the aws credentials in it")
        parser.add_argument("--half-life"
            , help = "Choose a half life for your new key"
            , choices = ["hour", "day", "week"]
            )
        if argv and argv[0] in ("--help", "-h"):
            self.args_from_subparser(action, parser, argv)
            # argparse should already quit before this point
            sys.exit(1)
        else:
            if not argv:
                raise NoExecCommand("argv is empty!")
            return {"command": argv}, do_exec

    def parse_register_saml(self, action, argv):
        """Args for registering a saml provider"""
        parser = argparse.ArgumentParser(description="Register a saml provider")

        parser.add_argument("--provider"
            , help = "The name of the provider"
            )

        args = self.args_from_subparser(action, parser, argv)
        return args, do_register_saml

    def parse_sourceable(self, action, argv):
        """Use CliParser to determine if these args given to credo produces a result that can be sourced into the shell"""
        parser = argparse.ArgumentParser(description="Entrypoint for scripts to determine if provided arguments, when given to credo, produces a result that should be sourced into the shell")
        if argv and argv[0] in ("--help", "-h"):
            try:
                self.args_from_subparser(action, parser, argv)
            except SystemExit:
                # Want to quit with an error code
                # So bash helper doesn't source the --help output
                sys.exit(1)
        else:
            if not argv:
                sys.exit(1)
            else:
                _, action, action_args = CliParser().split_argv(argv)
                if "--no-sourcing" not in action_args:
                    if action and action in self.actions:
                        parser = self.actions[action]
                        if getattr(parser, "sourceable", False):
                            sys.exit(0)
                        elif getattr(parser, "sourceable_if", None) == action:
                            sys.exit(0)

                # Otherwise, not sourceable
                sys.exit(1)

    def parse_serve(self, action, argv):
        """Args for registering a saml provider"""
        parser = argparse.ArgumentParser(description="Serve a fake metadata service")

        parser.add_argument("--port"
            , help = "The port to serve it under"
            , type = int
            , default = 80
            )

        parser.add_argument("--host"
            , help = "The host to serve it as"
            , default = "169.254.169.254"
            )

        args = self.args_from_subparser(action, parser, argv)
        return args, do_serve

    def parse_switch(self, action, argv):
        """Args for registering a saml provider"""
        parser = argparse.ArgumentParser(description="Switch what creds the metadata service returns")

        parser.add_argument("--port"
            , help = "The port to serve it under"
            , type = int
            , default = 80
            )

        parser.add_argument("--host"
            , help = "The host to serve it as"
            , default = "169.254.169.254"
            )

        args = self.args_from_subparser(action, parser, argv)
        return args, do_switch

def main(argv=None):
    __import__("boto")
    useragent = sys.modules["boto.connection"].UserAgent
    sys.modules["boto.connection"].UserAgent = "{0} Credo/{1}".format(useragent, VERSION)

    try:
        credo, kwargs, function = CliParser().parse_args(argv)
        function(credo, **kwargs)
    except CredoError as error:
        print("")
        print("!" * 80)
        print("Something went wrong! -- {0}".format(error.__class__.__name__))
        print("\t{0}".format(error))
        sys.exit(1)

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        pass




import re

class SamlRole(object):
    def __init__(self, principal_arn, role_arn):
        self.role_arn = role_arn
        self.principal_arn = principal_arn
        info = re.match("arn:aws:iam::(?P<account_id>[^:]+):role/(?P<role_name>.+)", self.role_arn).groups()
        self.account_id, self.role_name = info

    def __str__(self):
        return "{0} : {1}".format(self.account_id, self.role_name)

    def encrypted_values(self):
        return "{0},{1}".format(self.principal_arn, self.role_arn)

class SamlInfo(object):
    type = "saml"

    def __init__(self, provider, role, idp_username):
        self.role = role
        self.provider = provider
        self.idp_username = idp_username
        self.changed = True

    def needs_rotation(self):
        return False

    @property
    def access_keys(self):
        return []

    def unchanged(self):
        self.changed = False

    @property
    def encrypted_values(self):
        return {"provider": self.provider, "role": self.role.encrypted_values(), "idp_username": self.idp_username}, []




from credo.errors import BadKeyFile, CredoError, UserQuit
from credo.asker import ask_for_choice
from credo.structure.keys import Keys

import logging
import os

log = logging.getLogger("credo.cred_types.environment")

class EnvironmentMixin(object):
    requires_encryption = True

    @property
    def environment_location(self):
        """Where we should store environment vars"""
        return os.path.join(self.location, "env.json")

    def get_env_file(self, crypto):
        """Get us an EnvironmentFile object"""
        if not hasattr(self, "_env_file_cache"):
            self._env_file_cache = {}
        return EnvironmentFile.find_env(self.environment_location, crypto, self, cache=self._env_file_cache)

    def add_env(self, env, crypto):
        """Add another environment variable"""
        keys = self.get_env_file(crypto)

        for key, val in env:
            keys.add(key, val)

        keys.save()

    def remove_env(self, env, crypto):
        """Remove an environment variable"""
        keys = self.get_env_file(crypto)

        for key in env:
            keys.remove(key)

        keys.save()

    def shell_exports(self):
        """Get us some environment exports if there are any"""
        keys, error = EnvironmentFile.loaded_file_from(self.environment_location, self.crypto, self)
        if not keys:
            keys = EnvironmentFile(self.environment_location, self.crypto, self)

        return keys.shell_exports()

class EnvironmentFile(Keys):
    """Collection of environment variables"""
    def __init__(self, location, crypto, owner):
        self.owner = owner
        self.crypto = crypto
        self.location = location
        self._changed = False

    @property
    def changed(self):
        return self._changed

    def unchanged(self):
        """Set as not changed"""
        self._changed = False

    @classmethod
    def loaded_file_from(kls, location, crypto, owner):
        """Return the shell exports from this environment file and log errors"""
        if not os.path.exists(location):
            return None, False

        keys = EnvironmentFile(location, crypto, owner)
        try:
            keys.load()
        except CredoError as error:
            log.warning("Failed to load environment variables\tlocation=%s\terror_type=%s\terror=%s", location, error.__class__.__name__, error)
            return None, error

        return keys, None

    @classmethod
    def find_env(kls, location, crypto, owner, cache=None):
        """Add environment variables to our environment file"""
        if cache and location in cache:
            return cache[location]

        while True:
            keys, error = EnvironmentFile.loaded_file_from(location, crypto, owner)
            if error:
                quit_choice = "Quit"
                fixed_choice = "I fixed it, try again"
                override_choice = "Override the existing file"
                choice = ask_for_choice("Couldn't read the environment file, what do you want to do?", choices=[quit_choice, fixed_choice, override_choice])
                if choice == quit_choice:
                    raise UserQuit()
                elif choice == fixed_choice:
                    continue
                else:
                    break
            else:
                break

        if not keys:
            keys = EnvironmentFile(location, crypto, owner)

        if cache is not None:
            cache[location] = keys

        return keys

    @property
    def type(self):
        """This needs to be set"""
        return "environment"

    def add(self, key, value):
        """Add a key"""
        Empty = type("Empty", (object, ), {})
        if self.keys.get(key, Empty) != value:
            self._changed = True
            self.keys[key] = value

    def remove(self, key):
        """Remove a key"""
        Empty = type("Empty", (object, ), {})
        if self.keys.get(key, Empty) is not Empty:
            self._changed = True
            del self.keys[key]

    @property
    def encrypted_values(self):
        """Return our keys as a dictionary with encrypted values"""
        result = []
        log.info("Making encrypted values for environment variables")
        fingerprints = self.crypto.fingerprinted(self.keys)
        result = {"fingerprints": fingerprints}
        return result, self.keys.keys()

    def make_keys(self, contents):
        """Get us our keys from the contents of the file"""
        if contents.typ != self.type:
            raise BadKeyFile("Unknown type", type=contents.typ)
        keys = self.crypto.decrypt_by_fingerprint(contents.keys.get("fingerprints", {}), lambda *args, **kwargs: True)
        if not keys:
            return {}
        else:
            return keys

    def exports(self):
        """Return list of (key, val) exports we want to have in the shell"""
        return sorted(self.keys.items())

    def extra_env(self):
        """Return extra env stuff"""
        return getattr(self.owner, "extra_env", lambda: [])()

    @property
    def parent_path_part(self):
        """Return our parent path part"""
        return getattr(self.owner, "parent_path_part", None)

    @property
    def default_keys_type(self):
        """Empty keys is a dict"""
        return dict

    @property
    def default_keys_type_name(self):
        """Assume type is environment"""
        return "environment"

    def as_string(self):
        """Return information about keys as a string"""
        return "Environment vars! {0}".format(", ".join(self))




from __future__ import print_function

from credo.errors import BadConfigFile, BadSSHKey, BadCredentialSource, ProgrammerError, UserQuit

from six.moves import configparser, input
from collections import OrderedDict
from itertools import chain
import getpass
import logging
import boto
import sys
import os

log = logging.getLogger("credo.asker")

def get_response(*messages, **kwargs):
    """Get us a response from the user"""
    password = kwargs.get("password", False)
    if password:
        prompt = kwargs.get("prompt", ":")
    else:
        prompt = kwargs.get("prompt", ": ")

    for message in messages:
        if isinstance(message, dict):
            for num, val in message.items():
                print("{0}) {1}".format(num, val), file=sys.stderr)
        elif isinstance(message, list):
            for msg in message:
                print(msg, file=sys.stderr)
        else:
            print(message, file=sys.stderr)

    if prompt:
        sys.stderr.write(str(prompt))
        sys.stderr.flush()

    try:
        if password:
            return getpass.getpass(" ")
        else:
            return input() or kwargs.get("default", "")
    except KeyboardInterrupt:
        raise UserQuit()
    except EOFError:
        raise UserQuit()

def ask_for_choice(message, choices):
    """Ask for a value from some choices"""
    mapped = OrderedDict(enumerate(choices))
    no_value = True
    while no_value:
        response = get_response(message, "Please choose a value from the following", mapped)

        if response is None or not response.isdigit() or int(response) not in mapped:
            print("Please choose a valid response ({0} is not valid)".format(response), file=sys.stderr)
        else:
            no_value = False
            return mapped[int(response)]

def ask_for_choice_or_new(needed, choices):
    mapped = dict(zip(range(len(choices)), choices))
    while True:
        if mapped:
            maximum = max(mapped.keys())
            response = get_response(
                  "Choose a {0}".format(needed), "Please choose a value from the following"
                , mapped, {maximum+1: "Make your own value"}
                )

            if response is None or not response.isdigit() or int(response) < 0 or int(response) > maximum + 1:
                print("Please choose a valid response ({0} is not valid)".format(response), file=sys.stderr)
                continue
            else:
                response = int(response)
                if response in mapped:
                    return mapped[response]

            return get_response(prompt="Enter your custom value: ")
        else:
            return get_response("Choose a {0}".format(needed), prompt="Enter your custom value: ")

secret_sources = {
      "specified": "Specify your own value"
    , "aws_config": "Your awscli config file"
    , "boto_config": "Your boto config file"
    , "environment": "Your current environment"
    , "saml_provider": "One of your registered saml providers"
    }

def ask_user_for_secrets(credo, source=None):
    """Ask the user for access_key and secret_key"""
    typ = "amazon"
    choices = []
    access_key_name = "AWS_ACCESS_KEY_ID"
    secret_key_name = "AWS_SECRET_ACCESS_KEY"

    environment = os.environ

    if access_key_name in environment and secret_key_name in environment:
        choices.append(secret_sources["environment"])

    if os.path.exists(os.path.expanduser("~/.aws/config")):
        choices.append(secret_sources["aws_config"])

    if os.path.exists(os.path.expanduser("~/.boto")):
        choices.append(secret_sources["boto_config"])

    if credo.providers:
        choices.extend(["Saml provider '{0}'".format(provider) for provider in credo.providers])

    val = None
    if not source:
        if choices:
            val = ask_for_choice("Method of getting credentials", choices + [secret_sources["specified"]])
        else:
            val = secret_sources["specified"]
    else:
        if source not in secret_sources.keys() and source not in secret_sources.values():
            raise BadCredentialSource("Unknown credential source", source=source)

        if source in secret_sources:
            source = secret_sources[source]

        log.info("Getting credentials from %s", source)

    if secret_sources["specified"] in (val, source):
        access_key = get_response(prompt="Access key: ")
        secret_key = get_response(prompt="Secret key: ")

    elif secret_sources["environment"] in (val, source):
        if access_key_name not in environment or secret_key_name not in environment:
            raise BadCredentialSource("Couldn't find environment variables for {0} and {1}".format(access_key_name, secret_key_name))
        access_key = environment[access_key_name]
        secret_key = environment[secret_key_name]

    elif secret_sources["boto_config"] in (val, source) or secret_sources["aws_config"] in (val, source):
        parser = configparser.SafeConfigParser()
        aws_location = os.path.expanduser("~/.aws/config")
        boto_location = os.path.expanduser("~/.boto")

        if source == secret_sources["aws_config"] and not os.path.exists(aws_location):
            raise BadCredentialSource("Couldn't find the aws config", location=aws_location)
        if source == secret_sources["boto_config"] and not os.path.exists(boto_location):
            raise BadCredentialSource("Couldn't find the boto config", location=boto_location)

        if secret_sources["boto_config"] in (val, source):
            location = boto_location
        else:
            location = aws_location

        # Read it in
        parser.read(location)

        # Find possilbe sections
        sections = []
        for section in boto.config.sections():
            if section in ("Credentials", "default"):
                sections.append(section)

            elif section.startswith("profile "):
                sections.append(section)

        # Get sections that definitely have secrets
        sections_with_secrets = []
        for section in sections:
            if parser.has_option(section, "aws_access_key_id") and (parser.has_option(section, "aws_secret_access_key") or parser.has_option(section, "keyring")):
                sections_with_secrets.append(section)

        if not sections:
            raise BadConfigFile("No secrets to be found in the amazon config file", location=location)
        elif len(sections) == 1:
            section = sections[0]
        else:
            section = ask_for_choice("Which section to use?", sections)

        access_key = parser.get(section, "aws_access_key_id")
        if parser.has_option(section, "aws_secret_access_key"):
            secret_key = parser.get(section, "aws_secret_access_key")
        else:
            import keyring
            keyring_name = parser.get(section, 'keyring')
            secret_key = keyring.get_password(keyring_name, access_key)

    elif secret_sources["saml_provider"] in (val, source) or "Saml provider" in val:
        return "saml", "idp.realestate.com.au"

    else:
        raise ProgrammerError("Not possible to reach this point", source=source)

    return typ, (access_key, secret_key)

def ask_for_public_keys(remote=None, known_private_key_fingerprints=None):
    """
    Get keys from the user (urls, pems, locations)
    Where locations is a map of {<pem>: <location>} for when we know the location
    """
    urls = []
    pems = []
    locations = {}

    quit_choice = "Quit"
    no_choice = "No thanks"
    choice = no_choice

    if remote:
        if "@" in remote:
            _, path = remote.split("@", 1)
        if ":" in path:
            domain, path = path.split(":", 1)
        if "/" in path:
            username, path = path.split("/", 1)
        else:
            username = path

        suggestion = "https://{0}/{1}.keys".format(domain, username)
        suggestion_choice = "Use {0}".format(suggestion)
        choice = ask_for_choice_or_new("Do you want to use a url to get public keys for this repository?", [quit_choice, no_choice, suggestion_choice])

    while True:
        if choice == quit_choice:
            raise UserQuit()
        elif choice == no_choice:
            break
        elif choice == suggestion_choice:
            urls.append(suggestion)
        else:
            urls.append(choice)

        no_choice = "No more"
        choice = ask_for_choice_or_new("Do you to use any other urls for public keys?", [no_choice])

    public_key_pems = {}
    public_key_locations = {}
    public_key_fingerprints = {}

    from credo.crypto import KeyCollection
    collection = KeyCollection()
    ssh_folder = os.path.expanduser("~/.ssh")

    if os.path.exists(ssh_folder):
        available = list(chain.from_iterable(
            [ os.path.join(root, filename) for filename in files if filename.endswith(".pub")]
            for root, _, files in  os.walk(ssh_folder)
        ))

        for location in available:
            try:
                with open(location) as fle:
                    contents = fle.read()
                fingerprint = collection.add_public_key(contents)
                public_key_pems[location] = contents
                public_key_locations[fingerprint] = location
                public_key_fingerprints[location] = fingerprint
            except OSError as err:
                log.warning("Couldn't read %s (%s)", location, err)
            except BadSSHKey as err:
                log.warning("%s is not a valid public key (%s)", location)

    while True:
        question = "any"
        no_choice = "No thanks"
        quit_choice = "Quit"
        if pems:
            question = "any more"
            no_choice = "No more"

        public_key_fingerprints_choices = {}
        for location, fingerprint in public_key_fingerprints.items():
            public_key_fingerprints_choices["{0} ({1})".format(location, fingerprint)] = location

        if known_private_key_fingerprints:
            log.info("Know the following private keys\n\t%s", "\n\t".join(v.decode('utf-8') for v in known_private_key_fingerprints))
        choice = ask_for_choice_or_new("Do you want to add {0} public pem lines?".format(question), [quit_choice, no_choice] + sorted(public_key_fingerprints_choices.keys()))
        if choice == quit_choice:
            raise UserQuit()
        elif choice == no_choice:
            break
        elif choice in public_key_fingerprints_choices:
            choice = public_key_fingerprints_choices[choice]
            pem = public_key_pems[choice]
            locations[pem] = location
            pems.append(public_key_pems[choice])
            fingerprint = public_key_fingerprints[choice]
            del public_key_pems[choice]
            del public_key_fingerprints[choice]
            if fingerprint in public_key_locations:
                del public_key_locations[fingerprint]
        else:
            try:
                fingerprint = collection.add_public_key(choice)
                pems.append(choice)
                location = public_key_locations.get(fingerprint)
                if location:
                    del public_key_locations[fingerprint]
                    del public_key_fingerprints[location]
            except BadSSHKey as err:
                log.warning("The key you entered is not a valid public key (%s)", err)

    return urls, pems, locations

def ask_user_for_half_life(access_key):
    """Ask the user for a half life value"""
    day_choice = "One day"
    hour_choice = "One hour"
    week_choice = "One week"
    forever_choice = "forever"

    while True:
        choice = ask_for_choice_or_new("What half life do you want for this key? ({0})".format(access_key), [hour_choice, day_choice, week_choice, forever_choice])

        if choice == hour_choice:
            return 3600
        elif choice == day_choice:
            return 3600 * 24
        elif choice == week_choice:
            return 3600 * 24 * 7
        elif choice == forever_choice:
            return -1
        else:
            if not choice.isdigit():
                print("Please enter an integer representing the number of seconds in the half life", file=sys.stderr)
            else:
                return int(choice)

def ask_for_ssh_key_folders(already_have=None):
    """Ask for folders where we can find private keys"""
    if already_have is None:
        already_have = []
    originally_already_have = list(already_have)
    home_ssh = os.path.expanduser("~/.ssh")

    result = []
    while True:
        quit_choice = "Quit"
        choices = [quit_choice]

        retry_choice = None
        if originally_already_have:
            retry_choice = "I've added more keys to existing folders ({0}), try those again".format(", ".join(originally_already_have))
            choices.append(retry_choice)

        retry_home_choice = None
        if not os.path.exists(home_ssh):
            retry_home_choice = "I've used ssh-keygen, look for keys again"
            choices.append(retry_home_choice)

        if os.path.exists(home_ssh) and home_ssh not in already_have:
            choices.append(home_ssh)

        choice = ask_for_choice_or_new("Where can we find private ssh keys?", choices=choices)
        if choice == quit_choice:
            raise UserQuit()
        elif choice == retry_choice:
            break
        elif choice == retry_home_choice:
            if not os.path.exists(home_ssh):
                log.error("You say you've used ssh-keygen, but ~/.ssh (%s) doesn't exist....", home_ssh)
                continue
            else:
                if home_ssh not in result:
                    result.append(home_ssh)
                return result
        else:
            if not os.path.exists(choice):
                log.error("Provided folder doesn't exist!\tfolder=%s", choice)
            else:
                if choice in already_have:
                    log.info("Already have that folder!\tfolder=%s", choice)
                else:
                    result.append(choice)
                    already_have.append(choice)

        more_choice = "More folders"
        enough_choice = "No more folders"
        choice = ask_for_choice("What next?", choices=[quit_choice, more_choice, enough_choice])
        if choice == quit_choice:
            raise UserQuit()
        elif choice == enough_choice:
            break

    return result

def ask_for_env(part, env, remove_env, ask_for_more=False):
    """Ask the user for environment variables to store"""
    if env:
        part.add_env(env, part.crypto)
    if remove_env:
        part.remove_env(remove_env, part.crypto)

    if ask_for_more or not env:
        env_file = part.get_env_file(part.crypto)
        while True:
            stop_choice = "Stop specifying environment variables"
            show_choice = "Show what we currently have stored"
            add_new_choice = "Create new environment variable"
            uncapture_choice = "Uncapture a variable"
            use_existing_choice = "Use existing environment variable"

            choices = [stop_choice, show_choice, add_new_choice, use_existing_choice]
            if any(env_file.keys):
                choices.append(uncapture_choice)
                print("Captured env variables for [{0}]".format(", ".join('"{0}"'.format(key) for key in env_file.keys.keys())), file=sys.stderr)

            choice = ask_for_choice("What to do?", choices=choices)
            if choice == stop_choice:
                break
            elif choice == show_choice:
                print("", file=sys.stderr)
                for key, val in env_file.keys.items():
                    print("{0} = {1}".format(key, val), file=sys.stderr)
                    print("---", file=sys.stderr)
                    print("", file=sys.stderr)

                if hasattr(part, "parent_path_part"):
                    shell_exports = part.parent_path_part.shell_exports()
                    if shell_exports:
                        print("=" * 80)
                        print("Overriding")
                        print("-" * 40)
                        print("")
                        for key, val in shell_exports:
                            print("{0} = {1}".format(key, val), file=sys.stderr)
                            print("===", file=sys.stderr)
                            print("", file=sys.stderr)

            elif choice == add_new_choice:
                name = get_response(prompt="Name: ")
                value = get_response(prompt="Value: ")
                part.add_env([(name, value)], part.crypto)
            elif choice == use_existing_choice:
                name = get_response(prompt="Name: ")
                if name not in os.environ:
                    log.error("There is no environment variable called %s", name)
                else:
                    part.add_env([(name, os.environ[name])], part.crypto)
            elif choice == uncapture_choice:
                ignore_choice = "Actually, I don't want to uncapture any"
                choices = [ignore_choice] + env_file.keys.keys()
                choice = ask_for_choice("Which do you want to remove?", choices=choices)
                if choice != ignore_choice:
                    part.remove_env([choice], part.crypto)

    return env, remove_env

def ask_user_for_saml(credo, action=None, provider=None):
    """Ask the user for saml providers"""
    if action not in (None, "delete", "add", "show"):
        raise ProgrammerError("Given a saml action other than delete or add", got=action)

    first = True
    supplied_action = action
    while True:
        quit_action = "Quit"
        add_action = "Add a provider"
        show_action = "Show providers"
        delete_action = "Delete a provider"
        no_more_action = "No more"

        if supplied_action:
            action = {"add": add_action, "delete": delete_action, "show": show_action}[supplied_action]

        if not supplied_action:
            question = "Do you want to add a provider or remove a provider?"
            choices = [quit_action]
            if not first:
                choices.append(no_more_action)

            if credo.providers:
                question = "{0} or remove a provider?".format(question)
                choices.extend([add_action, show_action, delete_action])
            else:
                question = "{0}?".format(question)
                choices.append(add_action)

            action = ask_for_choice(question, choices)

        if action == no_more_action:
            return

        elif action == show_action:
            print("Current providers are:")
            for provider in credo.providers:
                print("\t{0}".format(provider))

        elif action == add_action:
            if not provider:
                provider = get_response("Saml provider")

            credo.register_saml_provider(provider)
            if supplied_action:
                return

        elif action == delete_action:
            if not credo.providers:
                log.info("No providers to delete!")
                if supplied_action:
                    return

            removing = provider
            if not removing:
                removing = ask_for_choice("Which provider do you want to remove?", choices=credo.providers)

            credo.remove_saml_provider(removing)
            if supplied_action:
                return

        elif action == quit_action:
            raise UserQuit()

        # Reset for next loop iteration
        first = False
        action = None
        provider = None
        supplied_action = False




class Base(object):
    """Everything a versioning driver needs to know"""
    def __init__(self, location):
        self.location = location

    @property
    def remote(self):
        """Return the address of our external location"""
        if not hasattr(self, "_remote"):
            self._remote = self.determine_remote()
        return self._remote

    @remote.setter
    def remote(self, val):
        """Record a new remote"""
        self._remote = val

    @property
    def versioned(self):
        """Say whether we are versioned"""
        return self.is_versioned()

    def is_versioned(self):
        """Boolean saying whether we are versioned"""
        raise NotImplementedError()

    def determine_remote(self):
        """Figure out what our remote is"""
        raise NotImplementedError()

    def add_change(self, message, change_files):
        """Add a change locally that gives specified message or changes in the specified files"""
        raise NotImplementedError()

    def synchronize(self, override=False):
        """Make sure we are in sync with our external location"""
        raise NotImplementedError()

    def deleteme(self):
        """Delete what is necessary to stop versioning"""
        raise NotImplementedError()

    def initialise(self, new_remote=None):
        """Initialise the versioning!"""
        raise NotImplementedError()

    def change_remote(self, new_remote):
        """Setup the new remote"""
        raise NotImplementedError()

class NoVersioningDriver(Base):
    """Basically no-ops everything"""
    def deleteme(self): pass
    def initialise(self, new_remote=None): pass
    def add_change(self, message, change_files): pass
    def synchronize(self, override=False): pass
    def is_versioned(self): return False
    def change_remote(self): pass
    def determine_remote(self): return None




from __future__ import print_function

from credo.asker import ask_for_choice, ask_for_public_keys
from credo.errors import BadConfiguration, UserQuit

import requests
import logging
import json
import time
import sys
import six
import os

log = logging.getLogger("credo.pub_keys")

class PubKeySyncer(object):
    """Knows about what public keys we can encrypt with"""
    def __init__(self, root_dir, repository):
        self.repository = repository
        self.crypto = self.repository.crypto
        self.root_dir = root_dir

    def sync(self, ask_anyway=False):
        """
        Find public keys for this repository and add them to the crypto object
        And remove public keys we don't know about anymore
        """
        info = {}
        added = []
        crypto = self.crypto

        first = True
        while first or (ask_anyway or not crypto.can_encrypt):
            first = False
            if "urls" not in info:
                info["urls"] = []
            if "pems" not in info:
                info["pems"] = []
            if "locations" not in info:
                info["locations"] = {}

            urls, pems, locations, new_ones = self.get_public_keys(ask_anyway=ask_anyway, known_private_key_fingerprints=self.crypto.private_key_fingerprints)
            if not new_ones and ask_anyway:
                break

            info["urls"].extend(urls)
            info["pems"].extend(pems)
            info["locations"].update(locations)

            downloaded = []
            for url in info.get("urls", []):
                downloaded.extend(self.download_pems(url))
            info["pems"].extend(downloaded)

            for pem in info["pems"]:
                location = info["locations"].get(pem)
                fingerprint = crypto.add_public_keys([(pem, location)]).get(pem)
                added.append(fingerprint)
                if not fingerprint:
                    log.error("Failed to add public key\tpem=%s", pem)
                else:
                    if location:
                        log.debug("Adding a public key\tlocation=%s\tfingerprint=%s", location, fingerprint)
                    else:
                        log.debug("Adding a public key\tfingerprint=%s", fingerprint)

            if not crypto.can_encrypt:
                log.error("Was unable to find any public keys")
                del info["urls"]
                del info["pems"]
            else:
                break

        for fingerprint in list(crypto.public_key_fingerprints):
            if fingerprint not in added:
                log.info("Removing public key we aren't encrypting with anymore\tfingerprint=%s", fingerprint)
                crypto.remove_public_key(fingerprint)

    def download_pems(self, url):
        """Get pems from some url"""
        cache = {}
        cache_location = os.path.join(self.root_dir, "cache")
        if os.path.exists(cache_location):
            try:
                with open(cache_location) as fle:
                    cache = json.load(fle)
            except ValueError as err:
                log.warning("Failed to load the pem url cache\tlocation=%s\terr=%s", cache_location, err)
        else:
            log.debug("No cache to load urls from\tlocation=%s", cache_location)

        last_downloaded = None
        if url in cache.get("cached", {}):
            if "times" in cache and url in cache["times"]:
                last_downloaded = cache["times"][url]
                if not isinstance(last_downloaded, float) and not isinstance(last_downloaded, int):
                    last_downloaded = None

                else:
                    diff = time.time() - last_downloaded
                    if diff < 0 or diff > 3600:
                        log.info("Cache for %s is older than an hour, re-getting the keys", url)
                        last_downloaded = None

        cached = []
        if last_downloaded is not None and url in cache.get("cached", {}):
            cached = cache["cached"][url]
            if isinstance(cached, list) and all(isinstance(pem, six.string_types) for pem in cached):
                log.info("Using %s cached pem keys from %s", len(cached), url)
                return cached

        if "times" not in cache:
            cache["times"] = {}
        cache["times"][url] = time.time()

        if "cached" not in cache:
            cache["cached"] = {}

        try:
            lines = requests.get(url).content.split('\n')
        except requests.exceptions.RequestException as err:
            lines = None
            log.error("Failed to get pem keys from url\turl=%s\terr=%s\treason=%s", url, err.__class__.__name__, err)
            if cached:
                log.info("Using %s keys from cache of %s", len(cached), url)

        if lines:
            cached = []
            for line in lines:
                if line.startswith("ssh-rsa"):
                    cached.append(line)

            cache["cached"][url] = cached
            log.info("Got %s keys from %s", len(cached), url)

            try:
                with open(cache_location, "w") as fle:
                    json.dump(cache, fle)
            except ValueError as err:
                log.error("Couldn't write cache\tlocation=%s\terr=e%s", cache_location, err)

        return cached

    def get_public_keys(self, ask_anyway=False, known_private_key_fingerprints=None, remote=None):
        """
        Return public keys for this repository as (urls, pems, locations, new_ones)
        Where locations is a map of {<pem>: <location>} for when we know the location
        and new_ones says whether we got any new ones from the user
        """
        keys_location = os.path.join(self.repository.location, "keys")

        result = {}
        locations = {}

        if os.path.exists(keys_location):
            try:
                with open(keys_location) as fle:
                    result = json.load(fle)
            except ValueError as err:
                result = self.fix_keys(keys_location, err)

            result = self.validate_keys_file(keys_location, result)

        new_ones = False
        if not os.path.exists(keys_location) or ask_anyway:
            urls, pems, locations = ask_for_public_keys(remote, known_private_key_fingerprints)
            self.crypto.add_public_keys(locations.items())

            if urls or pems:
                new_ones = True

            if "urls" not in result:
                result["urls"] = []
            if "pems" not in result:
                result["pems"] = []
            result["urls"].extend(urls)
            result["pems"].extend(pems)

        urls = result.get("urls", [])
        pems = result.get("pems", [])

        if urls or pems:
            for_signing = self.make_signing_value(urls, pems).encode('utf-8')
            result["signature"] = self.crypto.create_signature(for_signing)
            try:
                content = json.dumps(result, indent=4)
            except ValueError as err:
                raise BadConfiguration("Couldn't write out keys json", err=err)

            log.debug("Writing out public keys\tlocation=%s", keys_location)
            dirname = os.path.dirname(keys_location)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            with open(keys_location, 'w') as fle:
                fle.write(content)
            self.repository.add_change("Adjusting known public keys", [keys_location], repo=self.repository.name)

        return urls, pems, locations, new_ones

    def fix_keys(self, location, error):
        """Get user to fix the keys file"""
        info = {"error": error}
        while True:
            quit_choice = "Quit"
            remove_choice = "Remove the file"
            try_again_choice = "I fixed it, Try again"
            choices = [try_again_choice, remove_choice, quit_choice]
            response = ask_for_choice("Couldn't load {0} as a json file ({1})".format(location, info["error"]), choices)

            if response == quit_choice:
                raise UserQuit()

            elif response == remove_choice:
                os.remove(location)
                return {}

            else:
                try:
                    return json.load(location)
                except ValueError as err:
                    info["error"] = err

    def make_signing_value(self, urls, pems):
        """Make a signature for our public keys file"""
        return "urls:{0}||pems:{0}".format(",".join(sorted(urls)), ",".join(sorted(pems)))

    def validate_keys_file(self, location, contents):
        """Validate the signature in our contents file"""
        signature = contents.get("signature")
        if signature:
            if not isinstance(signature, list) or len(signature) != 2:
                log.warning("Signature for keys file is not a valid format\tlocation=%s", location)
                signature = None
        else:
            log.warning("Didn't find a signature in the keys file\tlocation=%s", location)

        if not signature:
            return self.manually_confirm_keys(location, contents)

        fingerprint, signature = signature
        for_signing = self.make_signing_value(contents.get("urls", []), contents.get("pems", []))

        has_public_key = self.crypto.retrieve_public_key_from_disk(fingerprint, reason="to check the signature in {0}".format(location))
        if has_public_key and self.crypto.is_signature_valid(for_signing, fingerprint, signature):
            return contents
        else:
            log.warning("Couldn't validate the signature for the keys file\tlocation=%s", location)
            return self.manually_confirm_keys(location, contents)

    def manually_confirm_keys(self, location, contents):
        """Manually confirm the urls and pems in our contents file"""
        result = {"pems": [], "urls": []}
        pems = contents.get("pems", [])
        urls = contents.get("urls", [])

        if pems:
            log.info("You need to confirm that you own the private keys for these public keys")
            fingerprints_with_pems = self.crypto.zip_with_fingerprints(pems)
            print("File contains public keys with these fingerprints", file=sys.stderr)
            print("\t{0}".format("\n\t".join(v[0].decode('utf-8') for v in fingerprints_with_pems)), file=sys.stderr)

            if self.crypto.private_key_fingerprints:
                print("\nWe know about these private keys", file=sys.stderr)
                print("\t{0}".format("\n\t".join(self.crypto.private_key_fingerprints)), file=sys.stderr)

            matches = all(fingerprint in self.crypto.private_key_fingerprints for fingerprint, _ in fingerprints_with_pems)
            matches_str = " (Not all the public keys match existing private keys)"
            if matches:
                matches_str = " (All the keys match private keys on your computer)"

            quit_choice = "Quit"
            confirm_all_choice = "I say all of them are valid {0}".format(matches_str)
            individual_confirm_choice = "I'll confirm them individually"

            if matches:
                confirm_all_choice = "I say all of them are valid (All the keys match private keys on this computer)"
            else:
                individual_confirm_choice = "I'll confirm them individually (Not all the keys match private keys on this computer)"

            choice = ask_for_choice("How do you want to confirm your keys?", [quit_choice, confirm_all_choice, individual_confirm_choice])

            if choice == quit_choice:
                raise UserQuit()
            elif choice == confirm_all_choice:
                result["pems"].extend(pems)
            else:
                for fingerprint, pem in fingerprints_with_pems:
                    matches = fingerprint in self.crypto.private_key_fingerprints
                    matches_str = ""
                    if matches:
                        matches_str = " (Matches a private key on this computer)"
                    quit_choice = "Quit"
                    exclude_choice = "No"
                    include_choice = "yes"
                    if matches:
                        include_choice = "Yes{0}".format(matches_str)
                    else:
                        exclude_choice = "No{0}".format(matches_str)

                    choice = ask_for_choice("Is {0} a fingerprint of yours?".format(fingerprint), [include_choice, exclude_choice, quit_choice])
                    if choice == quit_choice:
                        raise UserQuit()
                    elif choice == include_choice:
                        result["pems"].append(pem)

        if urls:
            log.info("You need to confirm you want keys from these urls")
            print("\t{0}".join("\n\t".join(urls)))

            quit_choice = "Quit"
            confirm_all_choice = "All of them are valid"
            individual_confirm_choice = "I'll confirm them individually"
            choice = ask_for_choice("How do you want to confirm the urls?", [quit_choice, confirm_all_choice, individual_confirm_choice])

            if choice == quit_choice:
                raise UserQuit()
            elif confirm_all_choice:
                result["urls"].extend(urls)
            else:
                for url in urls:
                    quit_choice = "Quit"
                    exclude_choice = "No"
                    include_choice = "Yes"
                    choice = ask_for_choice("Is this urls fine?", [include_choice, exclude_choice, quit_choice])
                    if choice == quit_choice:
                        raise UserQuit()
                    elif choice == include_choice:
                        result["urls"].append(url)

        return result




from credo.cred_types.saml import SamlInfo, SamlRole
from credo.cred_types.amazon import AmazonKeys
from credo.errors import BadCredentialFile
from credo.structure.keys import Keys
from credo.asker import get_response
from credo.amazon import IamSaml
import logging

log = logging.getLogger("credo.structure.credentials")

class Credentials(Keys):
    """Knows about credential files"""
    requires_encryption = True

    def save(self, force=False, half_life=None):
        """Save our credentials to file"""
        if self.keys.needs_rotation():
            self.keys.rotate(half_life)

        if force or self.changed or self.keys.changed:
            self.contents.save(self.location, self.keys, access_keys=list(self.keys.access_keys))
            self.keys.unchanged()

            cred_path = self.credential_path
            self.credential_path.add_change("Saving new keys", [self.location]
                , repo=cred_path.repository.name, account=cred_path.account.name, user=cred_path.user.name
                )

    def invalidate_creds(self):
        """Mark our creds as invalid"""
        cred_path = self.credential_path
        log.info("Marking the credentials as invalid\trepo=%s\taccount=%s\tuser=%s", cred_path.repository.name, cred_path.account.name, cred_path.user.name)
        self.keys.invalidate_all()

    @property
    def path(self):
        """Return the repo, account and user this represents"""
        return "repo={0}|account={1}|user={2}|{3}".format(self.repo_name, self.account_name, self.name, self.path_name)

    @property
    def path_name(self):
        """Return this part of the path"""
        return "Credentials"

    @property
    def parent_path_part(self):
        """Return our user"""
        return self.credential_path.user

    def make_keys(self, contents):
        """Get us some keys"""
        if contents.typ != "amazon":
            raise BadCredentialFile("Unknown credentials type", found=contents.typ, location=contents.location)
        return AmazonKeys(contents.keys, self.credential_path)

    def exports(self):
        """The exports specific to the credentials stored"""
        return self.keys.exports()

    @property
    def default_keys_type(self):
        """Empty keys is a list"""
        return list

    @property
    def default_keys_type_name(self):
        """Assume type is amazon"""
        return "amazon"

    def as_string(self):
        """Return information about keys as a string"""
        return "keys!"

class SamlCredentials(Credentials):
    """Knows about saml information"""

    requires_encryption = False

    @property
    def path(self):
        return "role={0}|provider={1}|username={2}".format(self.role, self.provider, self.idp_username)

    @property
    def default_keys_type(self):
        """Empty keys is a list"""
        return dict

    @property
    def default_keys_type_name(self):
        """Assume type is amazon"""
        return "saml"

    @property
    def path_name(self):
        """Return this part of the path"""
        return "Saml Credentials"

    def as_string(self):
        """Return information about keys as a string"""
        return "Saml things!"

    def exports(self):
        """Export some values"""
        password = get_response("Password for idp user {0}".format(self.keys.idp_username), password=True)
        pair = IamSaml(self.keys.provider, self.keys.idp_username, password)
        return pair.exports(self.keys.role)

    @property
    def role(self):
        return self.keys.role

    @property
    def provider(self):
        return self.keys.provider

    @property
    def idp_username(self):
        return self.keys.idp_username

    def make_keys(self, contents):
        """Get us some keys"""
        if contents.typ != "saml":
            raise BadCredentialFile("Unknown credentials type", found=contents.typ, location=contents.location)

        if not contents.keys:
            contents.keys["role"] = self.role.encrypted_values()
            contents.keys["provider"] = self.provider
            contents.keys["idp_username"] = self.idp_username
        return SamlInfo(contents.keys["provider"], SamlRole(*contents.keys["role"].split(",")), contents.keys["idp_username"])




from credo.structure.credential_path import CredentialPath
from credo.errors import CredoError, SamlNotAuthorized
from credo.amazon import IamSaml
from credo import explorer

from six.moves.urllib.parse import urlencode
from datetime import datetime, timedelta
from six.moves import http_cookies
from boto.utils import parse_ts
from uuid import uuid1
import requests
import logging
import pickle
import time
import pytz
import sys
import os
import re

log = logging.getLogger("credo.server")

class Server(object):
    def __init__(self, host, port, credo):
        self.host = host
        self.port = port
        self.credo = credo
        self.stored_keys = {}
        self.stored_assertions = {}

    def start(self):
        try:
            from tornado.httpserver import HTTPServer
            from tornado.wsgi import WSGIContainer
            from tornado.ioloop import IOLoop
        except ImportError:
            raise CredoError("Please pip install tornado")
        http_server = HTTPServer(WSGIContainer(self.app))
        http_server.listen(self.port, self.host)
        log.info("Starting server on http://%s:%s", self.host, self.port)
        IOLoop.instance().start()

    @property
    def credentials(self):
        return getattr(self, "_credentials", None)

    @credentials.setter
    def credentials(self, val):
        self._credentials = val

    @property
    def basic_auth(self):
        if datetime.utcnow() > getattr(self, "_basic_auth_time", datetime.utcnow()) + timedelta(hours=4):
            self._basic_auth = None
        return getattr(self, "_basic_auth", None)

    @basic_auth.setter
    def basic_auth(self, val):
        self._basic_auth = val
        self._basic_auth_time = datetime.utcnow()

    @property
    def keys(self):
        keys, _ = self.set_keys(self.credentials)
        return keys

    def set_keys(self, credentials, timeout=None):
        keys = self.stored_keys.get(credentials.path)
        if keys is not None:
            expiration = parse_ts(keys["Expiration"])
            if timeout:
                expiration = keys["LastUpdated"] + timeout
            if datetime.utcnow() > expiration:
                log.info("Keys expired, recreating them")
                keys = None

        if keys is None:
            log.info("Assuming role")
            pair = IamSaml(credentials.provider, credentials.idp_username, "")
            pair.basic_auth = self.basic_auth
            keys = pair.get_result(credentials.role).credentials.to_dict()

            _keys = {
                  "Code": "Success"
                , "LastUpdated": datetime.utcnow()
                , "AccessKeyId": keys["access_key"]
                , "SecretAccessKey": keys["secret_key"]
                , "Token": keys["session_token"]
                , "Expiration": keys["expiration"]
                }
            self.stored_keys[credentials.path] = _keys
            self.stored_assertions[credentials.path] = pair.assertion
        return self.stored_keys[credentials.path], self.stored_assertions[credentials.path]

    def get_cookies(self, credentials):
        s = requests.Session()
        cookies = []
        def add_cookies(resp):
            for name, header in resp.headers.items():
                if name.lower() == 'set-cookie' and header not in cookies:
                    cookie = http_cookies.SimpleCookie()
                    cookie.load(header)
                    for value in cookie.values():
                        url = "https://.amazon.com/"
                        if value['domain']:
                            url = "https://{0}{1}".format(value['domain'], value['path'])
                        cookie = {"url": url, "name": value.key, "value": value.value}

                        gmt = pytz.timezone("GMT")
                        if value['expires']:
                            date = datetime.strptime(value['expires'], "%a, %d-%b-%Y %H:%M:%S GMT")
                            gmtd = gmt.localize(date)
                            cookie["expirationDate"] = gmtd.astimezone(pytz.timezone("UTC"))
                        else:
                            cookie["expirationDate"] = datetime.utcnow() + timedelta(days=365)
                        cookie["expirationDate"] = time.mktime(cookie['expirationDate'].timetuple())
                        cookies.append(cookie)

        resp = s.get("https://{0}".format(credentials.provider))
        add_cookies(resp)

        nxt = "https://{0}{1}".format(credentials.provider, re.findall('href="[^"]+"', resp.text)[0][6:-1])
        resp2 = s.get(nxt, allow_redirects=False)
        add_cookies(resp2)

        keys, assertion = self.set_keys(credentials, timeout=timedelta(minutes=5))
        nxt = "https://signin.aws.amazon.com/saml"
        data2 = urlencode({"SAMLResponse": assertion, "name": "", "roleIndex": credentials.role.role_arn, "RelayState": ""})
        headers2 = {"Content-Type": "application/x-www-form-urlencoded", "Content-Length": len(data2)}
        final_resp = s.post(nxt, data=data2, headers=headers2, allow_redirects=False)
        add_cookies(final_resp)

        return cookies

    def find_credentials(self, account):
        directory_structure, shortened = explorer.find_repo_structure(self.credo.root_dir, levels=3)
        explorer.narrow(shortened, ["Repository", "Account", "User"], None, want_new=False, forced_vals=[None, account, None])
        chains = explorer.flatten(directory_structure, shortened, want_new=False)
        chosen = list(self.credo.credentials_from(directory_structure, chains, complain_if_missing=True))[0]
        chosen.keys
        return chosen

    @keys.setter
    def keys(self, val):
        self._keys = val

    @property
    def app(self):
        try:
            from flask import Flask
        except ImportError:
            raise CredoError("Please pip install flask")

        if getattr(self, "_app", None) is None:
            self._app = Flask("credo.server")
            self._app.secret_key = uuid1()
            self.register_routes(self._app)
        return self._app

    def register_routes(self, app):
        from flask import jsonify, abort, make_response, request, redirect, session

        @app.route('/', methods = ['GET'])
        def index():
            return 'latest'

        @app.route('/latest/', methods = ['GET'])
        def latest():
            return 'meta-data'

        @app.route('/latest/meta-data/', methods = ['GET'])
        def meta_data():
            return 'iam\nswitch'

        @app.route('/latest/meta-data/switch/', methods = ["GET"])
        def switch_get():
            return 'to'

        @app.route('/latest/meta-data/switch/to/', methods = ["GET"])
        def switch_list():
            directory_structure, shortened = explorer.find_repo_structure(self.credo.root_dir, levels=3)
            html = []
            html.append("<body>")
            for repository, values in shortened.items():
                html.append("<h1>Repository: {0}</h1>".format(repository))
                html.append("<ul>")

                for account, values in sorted(values.items()):
                    html.append('<li><a href="/latest/meta-data/switch/to/{0}">{0}</a></li>'.format(account))

                html.append("</ul>")
            html.append("</body>")
            return '\n'.join(html)

        @app.route('/latest/meta-data/switch/to/<account>', methods = ["GET"])
        def switch_to(account):
            credentials = self.find_credentials(account)

            try:
                cookies = self.get_cookies(credentials)
            except SamlNotAuthorized:
                return make_response(jsonify({"error": "NEED_AUTH"}), 500)

            response = {"Location": "https://console.aws.amazon.com/console?region=ap-southeast-2", "Cookies": cookies}
            return make_response(jsonify(response))

        @app.route('/latest/meta-data/switch/', methods = ["POST"])
        def switch():
            if not request.data:
                return make_response(jsonify({"error": "Need post data"}), 500)
            obj = pickle.loads(request.data)
            basic_auth = obj.get("basic_auth", self.basic_auth)
            credentials = obj["credentials"]

            if basic_auth is None:
                return make_response(jsonify({"error": "NEED_AUTH"}), 500)
            else:
                self.basic_auth = basic_auth
                self.credentials = credentials
                self.keys = None

                # keys is a property that actually gets the credentials
                try:
                    self.keys
                except SamlNotAuthorized:
                    return make_response(jsonify({"error": "BAD_PASSWORD"}), 500)
                return "success"

        @app.route('/latest/meta-data/iam/', methods = ['GET'])
        def iam():
            return 'security-credentials'

        @app.route('/latest/meta-data/iam/security-credentials/', methods = ['GET'])
        def security_credentials():
            return 'BaseIAMRole'

        @app.route('/latest/meta-data/iam/security-credentials/BaseIAMRole', methods = ['GET'])
        def base_iam_role():
            if self.credentials is None or self.basic_auth is None:
                return make_response(jsonify({"error": "DO SWITCH"}), 500)
            return jsonify(self.keys)

        @app.errorhandler(400)
        def not_found(error):
            return make_response(jsonify({'error': 'bad request'}), 400)

        @app.errorhandler(404)
        def not_found(error):
            return make_response(jsonify({'error': 'not found'}), 404)

        @app.errorhandler(500)
        def not_found(error):
            return make_response(jsonify({'error': 'internal server error'}), 500)




from credo.errors import CredoError
import logging
import os

log = logging.getLogger("credo.versioning")

try:
    import pygit2
except ImportError:
    pygit2 = None

def has_git_abilities():
    """Useful if anything wants to know if we can git"""
    return pygit2 is not None

def determine_driver(location, version_type=None):
    """Determine what versioning driver to use for some location"""
    if version_type == "git" or os.path.exists(os.path.join(location, ".git")):
        if pygit2 is not None:
            from credo.versioning.git import GitDriver
            return GitDriver(location)
        else:
            log.warning("Can't import pygit2, so not doing any git things")
            version_type = None

    if version_type is None:
        from credo.versioning.base import NoVersioningDriver
        return NoVersioningDriver(location)

    raise CredoError("Unknown versioning type", version_type=version_type)




from credo.cred_types.environment import EnvironmentMixin
from credo.helper import SignedValueFile

import logging
import os

log = logging.getLogger("credo.structure.user")

class User(EnvironmentMixin):
    def __init__(self, name, location, credential_path):
        self.name = name
        self.location = location
        self.credential_path = credential_path

    @property
    def crypto(self):
        """Proxy credential_path"""
        return self.credential_path.crypto

    @property
    def repo_name(self):
        """Proxy credential_path"""
        return self.credential_path.repository.name

    @property
    def account_name(self):
        """Proxy credential_path"""
        return self.credential_path.account.name

    @property
    def username_location(self):
        """Location of where our username is"""
        return os.path.join(self.location, "username")

    @property
    def path(self):
        """Return the repo, account and user this represents"""
        return "repo={0}|account={1}|user={2}".format(self.repo_name, self.account_name, self.name)

    @property
    def parent_path_part(self):
        """Return our account"""
        return self.credential_path.account

    def extra_env(self):
        """Define default env stuff"""
        return [("CREDO_CURRENT_USER", self.name)]

    def username(self, suggestion=None, iam_pair=None):
        """
        Get us the amazon username for this user
        use suggestion and iam_pair to assist the user in choosing a value if there is none
        """
        if hasattr(self, "_username"):
            return self._username
        else:
            def suggestions():
                """Get us suggestions from suggestion and iam_pair"""
                result = []
                if suggestion:
                    result.append(suggestion)
                if iam_pair:
                    try:
                        result.append(iam_pair.ask_amazon_for_username())
                    except:
                        pass
                return result

            info_location = self.username_location
            signed_value_file = SignedValueFile(info_location, self.crypto, dict(repo=self.repo_name, account=self.account_name, username=self.name))
            question = "How do you want to enter the username?\trepo={0}\taccount={1}\tusername={2}".format(self.repo_name, self.account_name, self.name)
            username, created = signed_value_file.retrieve("Username", question, suggestions)

            if created:
                self.credential_path.add_change("Writing username {0}".format(username), [info_location], repo=self.repo_name, account=self.account_name, username=self.name)

            self._username = username

        return self._username




from credo.errors import (
      BadSSHKey, BadCypherText, BadFolder, CredoError
    , BadPlainText, PasswordRequired, BadPrivateKey, BadPublicKey
    , NoSuchFingerPrint, CantFindPrivateKey, InvalidData
    )
from credo.asker import ask_for_choice, get_response, ask_for_ssh_key_folders

from Crypto.Cipher import PKCS1_OAEP, AES
from Crypto.PublicKey import RSA
from Crypto import Random

from binascii import hexlify, unhexlify
from base64 import b64decode, b64encode
from paramiko import Message
import tempfile
import paramiko
import logging
import random
import string
import json
import os
import re

log = logging.getLogger("credo.crypto")

class KeyCollection(object):
    """Keeps private and public keys"""
    def __init__(self):
        self.public_fingerprints = {}
        self.private_fingerprints = {}

        self.private_key_locations = {}
        self.fingerprint_to_location = {}

        self._location_to_public_rsaobj = {}
        self._location_to_private_rsaobj = {}

        self._location_to_fingerprint = {}

    def get_any_private_fingerprint(self, **info):
        """Get any of the private fingerprints"""
        fullfilled = [fingerprint for fingerprint, rsaobj in self.private_fingerprints.items() if rsaobj and fingerprint in self.public_fingerprints]
        if fullfilled:
            return fullfilled[0]

        # None already, let's get one that we have a public key for
        unfullfilled = [fingerprint for fingerprint, rsaobj in self.private_fingerprints.items() if fingerprint in self.public_fingerprints]
        if unfullfilled:
            return self.make_fingerprint(self.private_rsaobj_for(unfullfilled[0])).decode('utf-8')

        raise CantFindPrivateKey(**info)

    def add_public_key(self, pem_data, location=None):
        """Record this public key"""
        rsaobj = self.rsaobj_from_pem(pem_data)
        fingerprint = self.make_fingerprint(rsaobj).decode('utf-8')
        self.public_fingerprints[fingerprint] = rsaobj
        if location:
            self.fingerprint_to_location[fingerprint] = location
        return fingerprint

    def remove_public_key(self, fingerprint):
        """Remove this public key"""
        if fingerprint in self.public_fingerprints:
            del self.public_fingerprints[fingerprint]

    def add_private_key(self, location):
        """
        Record this private key

        Only get the fingerprint from the public key for now if it needs a password
        So we can delay getting the password till it's absolutely necessary
        """
        rsaobj = self.rsaobj_from_location(location, only_need_public=True)
        fingerprint = self.make_fingerprint(rsaobj).decode('utf-8')
        self.private_fingerprints[fingerprint] = None
        self.private_key_locations[fingerprint] = location
        self.fingerprint_to_location[fingerprint] = location
        return fingerprint

    def location_for_fingerprint(self, fingerprint):
        """Return the location of the fingerprint if we have one"""
        return self.fingerprint_to_location.get(fingerprint)

    def public_pem_for(self, fingerprint):
        """Get the public pem data for this fingerprint"""
        return "ssh-rsa {0}".format(self.public_rsaobj_for(fingerprint).get_base64())

    def public_rsaobj_for(self, fingerprint):
        """Get the rsaobj for this public fingerprint"""
        if fingerprint in self.public_fingerprints:
            return self.public_fingerprints[fingerprint]

        if fingerprint in self.private_fingerprints:
            return self.private_fingerprints[fingerprint]

        raise NoSuchFingerPrint(fingerprint=fingerprint, looking_for="public")

    def private_rsaobj_for(self, fingerprint):
        """
        Get the rsaobj for this private fingerprint

        If we don't have the rsaobj but we do have the location,
        then we get the rsaobj for the private part as well first
        """
        if not self.private_fingerprints.get(fingerprint):
            if fingerprint not in self.private_key_locations:
                raise NoSuchFingerPrint(fingerprint=fingerprint, looking_for="private")

            rsaobj = self.rsaobj_from_location(self.private_key_locations[fingerprint])
            self.private_fingerprints[fingerprint] = rsaobj

        return self.private_fingerprints[fingerprint]

    def make_fingerprint(self, rsa_obj):
        """Get us a fingerprint from this rsa_obj"""
        st = hexlify(rsa_obj.get_fingerprint())
        return ":".join(re.findall("..", st.decode('utf-8'))).encode('utf-8')

    def rsaobj_from_pem(self, pem_data):
        """Get us a paramiko.RSAKey from a public key pem_data."""
        tmp = None
        try:
            tmp = tempfile.NamedTemporaryFile(delete=True).name
            with open(tmp, 'w') as fle:
                fle.write(pem_data)
            return self.make_rsakey(tmp)
        finally:
            if tmp and os.path.exists(tmp):
                os.remove(tmp)

    def rsaobj_from_location(self, location, only_need_public=False):
        """
        Get us a paramiko.RSAKey from this location

        If the location is a password protected private key and only_need_public then we look for a public key
        If we can't find a public key, then we ask for the password and get the fingerprint that way

        If it isn't a private key, then we raise a credo.NotSSHKey exception
        """
        if only_need_public:
            if location in self._location_to_public_rsaobj:
                return self._location_to_public_rsaobj[location]
        else:
            if location in self._location_to_private_rsaobj:
                return self._location_to_private_rsaobj[location]

        rsaobj = None
        try:
            rsaobj = self.make_rsakey(location, private=True)
            self._location_to_private_rsaobj[location] = rsaobj
        except PasswordRequired:
            if only_need_public:
                pub_key = "{0}.pub".format(location)
                if os.path.exists(pub_key):
                    try:
                        rsaobj = self.make_rsakey(pub_key)
                        self._location_to_public_rsaobj[location] = rsaobj
                    except BadSSHKey as err:
                        log.info("Something wrong with public key %s: %s", pub_key, err)

        if rsaobj is None:
            while True:
                if only_need_public:
                    log.info("Couldn't find a public key for password protected private key at %s", location)

                password = get_response("Password for your private key ({0})".format(location), password=True)

                try:
                    rsaobj = self.make_rsakey(location, password=password, private=True)
                    self._location_to_private_rsaobj[location] = rsaobj
                    break
                except BadSSHKey:
                    choice = ask_for_choice("Couldn't decode the key ({0})".format(location), ["Try again", "Ignore"])
                    if choice == "Ignore":
                        return

        return rsaobj

    def make_rsakey(self, location, password=None, private=False):
        """Get us an rsa object for this location"""
        try:
            if private:
                return paramiko.RSAKey.from_private_key_file(location, password=password)
            else:
                txt = open(location).read()
                if not txt.startswith("ssh-rsa"):
                    raise BadPublicKey("Doesn't start with ssh-rsa")
                split = txt.split(" ")
                if len(split) < 2:
                    raise BadPublicKey("Expecting more")

                try:
                    der = b64decode(split[1])
                except TypeError:
                    raise BadPublicKey("Couldn't decode")

                return paramiko.RSAKey(msg=Message(der))
        except paramiko.PasswordRequiredException:
            raise PasswordRequired()
        except paramiko.ssh_exception.SSHException as err:
            raise BadSSHKey("Couldn't decode key, perhaps bad password?", err=err)

class SSHKeys(object):
    """Stores private and public ssh keys by fingerprint"""
    def __init__(self):
        self._RSA = {}
        self.collection = KeyCollection()

    def have_private(self, fingerprint):
        """Says whether we have a private key with this fingerprint"""
        return fingerprint in self.collection.private_fingerprints

    def have_public(self, fingerprint):
        """Says whether we have a public key with this fingerprint"""
        return fingerprint in self.collection.public_fingerprints

    def find_keys(self, folder):
        """Find more private and public keys in specified folder"""
        if not os.path.exists(folder):
            raise BadFolder("Doesn't exist", folder=folder)
        if not os.access(folder, os.R_OK):
            raise BadFolder("Not readable", folder=folder)

        for root, dirs, files in os.walk(folder):
            for filename in files:
                location = os.path.join(root, filename)
                if not filename.endswith(".pub") and filename not in ("known_hosts", "authorized_keys", "config"):
                    if os.access(location, os.R_OK):
                        is_private = False
                        try:
                            self.collection.add_private_key(location)
                            is_private = True
                        except BadSSHKey:
                            pass

                        if not is_private:
                            try:
                                with open(location) as fle:
                                    self.collection.add_public_key(fle.read(), location)
                            except BadSSHKey:
                                pass

                elif filename.endswith(".pub"):
                    try:
                        with open(location) as fle:
                            self.collection.add_public_key(fle.read(), location)
                    except BadSSHKey:
                        pass

    def add_public_keys(self, public_keys):
        """Add the specified public keys"""
        fingerprints = {}
        for pem_data in public_keys:
            try:
                location = None
                if (isinstance(pem_data, tuple) or isinstance(pem_data, list)) and len(pem_data) == 2:
                    pem_data, location = pem_data
                fingerprints[pem_data] = self.collection.add_public_key(pem_data, location=location)
            except BadSSHKey as err:
                log.error("Found a bad public key\terr=%s", err)
        return fingerprints

    def remove_public_key(self, fingerprint):
        """Remove a public key from our collection"""
        self.collection.remove_public_key(fingerprint)

    def private_key_to_rsa_object(self, fingerprint, **info):
        """Get us a RSA object from our private key on disk"""
        if fingerprint in self._RSA:
            return self._RSA[fingerprint]

        if not self.have_private(fingerprint):
            raise BadPrivateKey("Don't have a private key for specified fingerprint", fingerprint=fingerprint)

        key = self.collection.private_rsaobj_for(fingerprint)
        location = self.collection.location_for_fingerprint(fingerprint)
        location_str = ""
        if location:
            location_str = "at {0} ".format(location)
        log.debug("Using private key %s(%s) to decrypt (%s)", location_str, fingerprint, " || ".join("{0}={1}".format(key, val) for key, val in info.items()))

        key = RSA.construct((key.n, key.e, key.d, key.p, key.q))
        self._RSA[fingerprint] = key
        return key

    def encrypt(self, message, fingerprint, **info):
        """Encrypt the specified message using specified public key and return as base64 encoded string"""
        log.debug("Using public key with fingerprint %s to encrypt (%s)", fingerprint, " || ".join("{0}={1}".format(key, val) for key, val in info.items()))
        rsakey = RSA.importKey(self.collection.public_pem_for(fingerprint))
        rsakey = PKCS1_OAEP.new(rsakey)
        try:
            encrypted = rsakey.encrypt(message)
        except ValueError as err:
            raise BadPlainText(error=err, **info)
        return b64encode(encrypted).decode('utf-8')

    def decrypt(self, package, fingerprint, **info):
        """Decrypt the specified base64 encoded package using specified private key"""
        key = self.private_key_to_rsa_object(fingerprint, **info)
        rsakey = PKCS1_OAEP.new(key)

        try:
            decoded = b64decode(package)
        except TypeError:
            raise BadCypherText("Value not valid base64 encoding", **info)

        try:
            return rsakey.decrypt(decoded)
        except ValueError as err:
            raise BadCypherText(err=err, **info)

class Crypto(object):
    """Knows how to do crypto"""
    def __init__(self, keys=None):
        if keys is None:
            keys = SSHKeys()
        self.keys = keys
        self.ssh_key_folders = []

    @property
    def can_encrypt(self):
        """Say whether we have any public keys"""
        return self.has_public_keys()

    @property
    def can_sign(self):
        """Say whether we have private keys with corresponding public keys"""
        return self.has_public_keys() and any(fingerprint in self.public_key_fingerprints for fingerprint in self.private_key_fingerprints)

    @property
    def private_key_fingerprints(self):
        """Proxy our key collection"""
        return self.keys.collection.private_fingerprints.keys()

    @property
    def public_key_fingerprints(self):
        """Proxy our key collection"""
        return self.keys.collection.public_fingerprints.keys()

    def has_public_keys(self):
        """Return whether we have any public keys"""
        return len(self.public_key_fingerprints) > 0

    def has_private_keys(self):
        """Return whether we have any private keys"""
        return len(self.private_key_fingerprints) > 0

    def find_keys(self, folder):
        """Find keys to add"""
        if folder not in self.ssh_key_folders:
            self.ssh_key_folders.append(folder)
        return self.keys.find_keys(folder)

    def add_public_keys(self, public_keys):
        """Add public keys"""
        return self.keys.add_public_keys(public_keys)

    def remove_public_key(self, fingerprint):
        """Remove a public key"""
        self.keys.remove_public_key(fingerprint)

    def decryptable(self, fingerprints):
        """Say whether we have a private key for any of these fingerprints"""
        return any(fingerprint in self.keys.collection.private_fingerprints for fingerprint in fingerprints)

    def is_signature_valid(self, signed, fingerprint, signature):
        """Return whether this signature is valid for the signed data"""
        try:
            return self.keys.collection.public_rsaobj_for(fingerprint).verify_ssh_sig(signed, paramiko.Message(unhexlify(signature)))
        except TypeError:
            return False

    def create_signature(self, for_signing):
        """Return a signature given this data"""
        fingerprint = self.keys.collection.get_any_private_fingerprint(need_private_key_for="signing")
        message = self.keys.collection.private_rsaobj_for(fingerprint).sign_ssh_data(for_signing)
        return fingerprint, hexlify(message.asbytes()).decode('utf-8')

    def zip_with_fingerprints(self, pems):
        """Return (fingerprint, pem) for each pem in pems"""
        result = []
        for pem in pems:
            try:
                rsaobj = self.keys.collection.rsaobj_from_pem(pem)
                fingerprint = self.keys.collection.make_fingerprint(rsaobj)
            except BadSSHKey as error:
                log.warning("Found a pem key that was invalid\terror=%s", error)

            result.append((fingerprint, pem))
        return result

    def retrieve_public_key_from_disk(self, fingerprint, reason=None):
        """Get this fingerprint and return whether we successfully did so"""
        if self.keys.collection.location_for_fingerprint(fingerprint):
            return True

        def make_ssh_key_location(folders):
            """Get a string to say where we have ssh key folders"""
            if len(self.ssh_key_folders) == 1:
                return self.ssh_key_folders[0]
            else:
                return "one of {0}".format(", ".join(self.ssh_key_folders))

        def add_more_ssh_key_folders():
            """Add more ssh key folders"""
            more_ssh_key_folders = ask_for_ssh_key_folders(already_have=self.ssh_key_folders)
            self.ssh_key_folders.extend(more_ssh_key_folders)

        while not self.ssh_key_folders:
            add_more_ssh_key_folders()
        ssh_key_locations = make_ssh_key_location(self.ssh_key_folders)

        while True:
            if self.keys.collection.location_for_fingerprint(fingerprint):
                return True

            ignore_choice = "I don't have this key"
            try_again_choice = "I've added the key to {0}".format(ssh_key_locations)
            have_another_ssh_key_folder = "I have another folder with ssh keys in it"
            choice = ask_for_choice("Looking for a public key with fingerprint {0}".format(fingerprint), choices=[ignore_choice, try_again_choice, have_another_ssh_key_folder])
            if choice == ignore_choice:
                return False
            elif choice == try_again_choice:
                continue
            elif choice == have_another_ssh_key_folder:
                add_more_ssh_key_folders()
                ssh_key_locations = make_ssh_key_location(self.ssh_key_folders)

            self.keys.find_public_keys()

    def decrypt_by_fingerprint(self, fingerprints, verifier, **info):
        """Return the first valid decrypted value"""
        for fingerprint, values in fingerprints.items():
            if self.keys.have_private(fingerprint) and set(["secret", "data", "verifier"]) - set(values.keys()) == set():
                verifier_val = values['verifier']
                if not isinstance(verifier_val, list) or len(verifier_val) != 2:
                    log.error("Ignoring value with invalid verifier (Verifier is not a list of [fingerprint, signature])")
                    continue

                fingerprint, signature = verifier_val
                encrypted_data = values['data']
                encrypted_secret = values['secret']

                secret = self.keys.decrypt(encrypted_secret, fingerprint, key_fingerprint=fingerprint, action="decrypting", value="Secret for decrypting with")
                decrypted_data = self.decrypt_with_secret(encrypted_data, secret)

                decrypted = None
                try:
                    decrypted = json.loads(decrypted_data)
                except (ValueError, TypeError) as error:
                    log.error("Couldn't load decrypted data as a json dictionary\terror_type=%s\terror=%s\tfingerprint=%s", error.__class__.__name__, error, fingerprint)

                if decrypted:
                    if not self.is_signature_valid(secret, fingerprint, signature):
                        log.error("Ignoring decrypted secrets, because can't verify signature\tfingerprint=%s", fingerprint)
                    else:
                        if not verifier(decrypted):
                            log.error("Ignoring invalid data")
                        else:
                            return decrypted

    def fingerprinted(self, decrypted_vals, **info):
        """
        Return dictionary of {<fingerprint>: {secret: <secret>, data:<data>, verifier:<verifier>}

        Where <secret> is a randomly generated secret
        <data> is the original data encrypted with AES using that secret
        and <verifier> is a signature that says the secret was created using this private key

        Decrypted_vals is assumed to be a json dictionary
        """
        if not isinstance(decrypted_vals, dict):
            raise CredoError("Fingerprinted should only be called with dictionaries", got_type=type(decrypted_vals))

        try:
            data_str = json.dumps(decrypted_vals, sort_keys=True)
        except (ValueError, TypeError) as error:
            raise InvalidData("Couldn't dump values for encryption", error_type=error.__class__.__name__, error=error, **info)

        result = {}
        for fingerprint in self.public_key_fingerprints:
            secret = self.generate_secret()
            verifier = self.create_signature(secret)
            encrypted_data = self.encrypt_with_secret(data_str, secret)
            log.info("Encrypting credentials using AES\tfingerprint=%s", fingerprint)
            encrypted_secret = self.keys.encrypt(secret, fingerprint, key_fingerprint=fingerprint, action="encrypting", value="Secret for encrypting with")
            result[fingerprint] = dict(secret=encrypted_secret, verifier=verifier, data=encrypted_data)

        return result

    def generate_secret(self, key_size=256):
        """Generate a secret that may be used for encrypting values"""
        return Random.OSRNG.posix.new().read(key_size // 8)

    def encrypt_with_secret(self, data, secret):
        """Return the data as an encrypted value, using AES with the provided secret"""
        def pad(s):
            x = AES.block_size - len(s) % AES.block_size
            return s + ''.join([random.choice(string.ascii_letters + string.digits) for n in range(x)])

        padded_message = pad(data)
        iv = Random.OSRNG.posix.new().read(AES.block_size)
        cipher = AES.new(secret, AES.MODE_CBC, iv)
        return b64encode(iv + cipher.encrypt(padded_message)).decode('utf-8')

    def decrypt_with_secret(self, ciphertext, secret):
        """Return the decrypted value of the ciphtertext using AES with the provided secret"""
        unpad = lambda s: s[:s.rfind("}")+1]
        decoded = b64decode(ciphertext)
        iv = decoded[:AES.block_size]
        cipher = AES.new(secret, AES.MODE_CBC, iv)
        return unpad(cipher.decrypt(decoded)[AES.block_size:].decode('utf-8'))




from credo.helper import copy_dict_structure

from fnmatch import fnmatch
import os

def filtered(original, looking_for=None, required_files=None):
    """
    Filter out unwanted entries from a nested dictionary structure

    Where looking_for is a list of strings specifying value at that level.

    So if looking_for is ["repo", None, "user1"] then we will return only those
    that have "repo" at level0 and "user1" at level2 and those required_files at that level
    """
    result = copy_dict_structure(original)
    if not looking_for or (not required_files and not any(looking_for)):
        return result

    def delete_not(original, last_level, wanted, only_if=None):
        """
        Delete from the bottom if the last level is not our wanted level

        if only_if is specified then delete if only_if returns False

        Delete up if what we delete has no siblings
        """
        thing = original
        if isinstance(thing, dict):
            if last_level > 0:
                for key, val in list(thing.items()):
                    delete_not(thing[key], last_level-1, wanted, only_if=only_if)
                    if not val or (last_level-1 > 0 and not isinstance(val, dict)):
                        del thing[key]
            else:
                for key, val in list(thing.items()):
                    if wanted and not fnmatch(key, wanted):
                        del thing[key]
                    elif only_if and not only_if(val):
                        del thing[key]

    def has_required_files(val):
        """
        Say if the required files are here.
        True if we have no required Files
        False if the value isn't a list
        False if all the required files aren't in the value
        True if the val is a list and all the required files are in it
        """
        if not required_files:
            return True
        if not isinstance(val, list):
            return False
        return all(required_file in val for required_file in required_files)

    last_level = len(looking_for) - 1
    wanted = looking_for[last_level]
    if wanted or required_files:
        delete_not(result, last_level, wanted, only_if=has_required_files)

    while last_level > -1:
        if last_level > -1:
            wanted = looking_for[last_level]
            if wanted:
                delete_not(result, last_level, wanted)
            last_level -= 1

    return result

def find_repo_structure(root_dir, collection=None, sofar=None, shortened=None, levels=3):
    """
    Recursively explore a directory structure and put it in a dictionary to the number of levels specified

    So say we had a directory of

        <root>/

            github.com:blah/
                prod/
                    .git/
                        HEAD
                        index/
                    user1/
                        credentials.json

            bitbucket.com:blah
                dev/
                    user1/
                        ignored/
                            ignored2/

    And we did

        collection, shortened = find_repo_structure(<root>)

    collection would become

        { "github.com:blah" :
            { "prod":
            { "user1": { "/files/": ["credentials.json"], "/dirs/": [], "/location/": <location> }
                , "/dirs/": []
            , "/files/": []
            , "/location/": "#{<root>}/github.com:blah/prod/"
            }
            , "/dirs/": [".git"]
            , "/files/": []
            , "/location/": "#{<root>}/github.com:blah/"
            }

        , "bitbucket.com:blah":
            { "dev":
                { "user1": {"/files/": [], "/dirs/": ["ignored"], "/location/": <location> }
                , "/dirs/": []
                , "/files/" []
                , "/location/": "#{<root>}/bitbucket.com:blah/dev/"
                }
            , "/dirs/": []
            , "/files/": []
            , "/location/": "#{<root>}/bitbucket.com:blah/"
            }
        , "/files/": []
        , "/location/": "#{<root>}"
        }

    and shortened would become

        {"github.com:blah": {"prod": {"user1": ["credentials.json"]}}}
    """
    dirs = []
    basenames = []
    extra_dirs = []

    sofar = [] if sofar is None else sofar
    shortened = {} if shortened is None else shortened
    collection = {} if collection is None else collection

    for filename in os.listdir(root_dir):
        location = os.path.join(root_dir, filename)
        if os.path.isfile(location):
            basenames.append(filename)
        else:
            if filename.startswith("."):
                extra_dirs.append(filename)
            else:
                dirs.append((filename, location))

    collection["/dirs/"] = extra_dirs
    collection["/files/"] = basenames
    collection["/location/"] = root_dir
    if not collection["/location/"].endswith('/'):
        collection["/location/"] = "{0}/".format(collection["/location/"])

    if levels == 0:
        s = shortened
        collection["/dirs/"] = extra_dirs + [dr for dr, _ in dirs]
        for part in sofar[:-1]:
            if part not in s:
                s[part] = {}
            s = s[part]
        s[sofar[-1]] = basenames
    else:
        for filename, location in dirs:
            nxt_collection = {}
            collection[filename] = nxt_collection
            find_repo_structure(location, collection=nxt_collection, sofar=list(sofar) + [filename], shortened=shortened, levels=levels-1)

    return collection, shortened

class Stop(object):
    """Used to stop searching in narrow"""

def narrow(structure, chain, asker, want_new=None, want_any_after=None, forced_vals=None, level=0):
    """Narrow down our mask to a single result"""
    if not chain:
        return
    else:
        nxt = chain.pop(0)
        if nxt is Stop:
            chosen = nxt
        else:
            chosen = None

        forced = False
        if forced_vals:
            chosen = forced_vals.pop(0)
            if chosen is not None:
                forced = True

        if not chosen and not forced:
            if want_any_after is not None and level >= want_any_after:
                choices = sorted(structure.keys())
                all_choice = "All {0}".format(nxt)
                chosen = asker(nxt, choices + [all_choice])
                if chosen == all_choice:
                    chosen = Stop
            else:
                if len(structure.keys()) > 1 or want_new:
                    chosen = asker(nxt, sorted(structure.keys()))
                elif structure:
                    chosen = list(structure.keys())[0]

        for key in list(structure.keys()):
            if chosen is not None and key != chosen:
                del structure[key]

        if not structure and want_new:
            structure[chosen] = {} if chain else []

        if structure:
            narrow(list(structure.values())[0], chain, asker, want_new=want_new, want_any_after=want_any_after, forced_vals=forced_vals, level=level+1)

def flatten(directory_structure, mask, want_new=False):
    """
    Given a collection and shortened like what find_repo_structure gives,
    return a list of lists of [final_location, <val>, <val>, ...]

    So given a directory_structure of:

        { "/location/": <root>
        , "/files/": <root_files>
        , "/dirs/": <root_extra_dirs>
        , "repo1":
            { "/location/" and "/files/" and "/dirs/"
            , "account1":
              { "/location/" and "/files/" and "/dirs/"
              , "user1":
                { <you get the idea>
                }
              , <other users>
              }
            , <other accounts>
            }
        , <other_repos>
        }

    And a mask of {"repo1": {"account1": {"user1": [], <other_user>: []}}}

    Return

        [ [<user1_root>, "repo1", "account1", "user1"]
        , [<other_user_root>, "repo1", "account1", <other_user>]
        ]

    """
    def fill_out(structure, mask, collected=None):
        """Recursively fill out the collected from structure"""
        new_collected = []

        for nxt, remaining_mask in mask.items():
            if nxt is None:
                break

            if nxt not in structure and want_new:
                location = os.path.join(structure['/location/'], nxt)
                structure[nxt] = {'/location/': location, '/files/': [], '/dirs/': []}

            if nxt in structure:
                if collected is None:
                    collected = [[structure["/location/"]]]
                extended = [collection + [nxt] for collection in collected]
                if isinstance(remaining_mask, dict):
                    new_collected.extend(fill_out(structure[nxt], remaining_mask, extended))
                else:
                    new_collected.extend(extended)

        return new_collected

    collected = fill_out(directory_structure, mask)
    return [[os.path.join(*chain)] + chain[1:] for chain in collected]




from setuptools import setup, find_packages
from credo import VERSION

setup(
      name = "credo_manager"
    , version = VERSION
    , packages = ['credo'] + ['credo.%s' % pkg for pkg in find_packages('credo')]
    , include_package_data = True

    , install_requires =
      [ "rainbow_logging_handler==2.2.0"
      , "pycrypto==2.6.1"
      , "paramiko==1.14.0"
      , "requests==2.8.1"
      , "keyring==3.5"
      , "boto>=2.32.1"
      , "delfick_error==1.6"
      , "pytz"
      , "six"
      ]

    , extras_require =
      { "tests":
        [ "noseOfYeti>=1.5.0"
        , "nose"
        , "mock"
        ]

      , "git":
        [ "pygit2"
        ]
      }

    , entry_points =
      { 'console_scripts' :
        [ 'credo = credo.executor:main'
        ]
      }

    # metadata for upload to PyPI
    , url = "https://github.com/delfick/credo"
    , author = "Stephen Moore"
    , author_email = "stephen@delfick.com"
    , description = "Manager for aws credentials"
    , license = "MIT"
    , keywords = "iam amazon credentials"
    )



from credo.asker import ask_for_choice_or_new, ask_user_for_half_life
from credo.errors import NoValueEntered, BadKeyFile, CredoError

import logging
import copy
import json
import six
import os

log = logging.getLogger("credo.helper")

def record_non_dicts(subject, memo):
    """Find all the things in subject that are not string or dicts and fill memo with {id(thing): thing}"""
    for key, val in subject.items():
        if not isinstance(val, dict) and not isinstance(val, six.string_types):
            memo[id(val)] = val

        elif isinstance(val, dict):
            record_non_dicts(val, memo)

def copy_dict_structure(structure):
    """Copy the structure of a dictionary without modifying any of the contents"""
    memo = {}
    record_non_dicts(structure, memo)
    return copy.deepcopy(structure, memo=memo)

def print_list_of_tuples(lst, prefix):
    """Helper for printing out a list of tuples"""
    if any(val for _, val in lst):
        print("{0}: {1}".format(prefix, " | ".join("{0}={1}".format(key, val) for key, val in lst if val)))

def make_export_commands(exports, no_transform=False):
    for key, val in exports:
        if val == "CREDO_UNSET" and not no_transform:
            yield "unset {0}".format(key)
        else:
            yield "export {0}=\"{1}\"".format(key, val.replace("\\\"", "\"").replace("\"", "\\\""))

def normalise_half_life(half_life, access_key=None):
    if half_life is None:
        if access_key is None:
            return
        half_life = ask_user_for_half_life(access_key)

    if isinstance(half_life, six.string_types) and not half_life.isdigit():
        if half_life == "hour":
            half_life = 60 * 60
        elif half_life == "day":
            half_life = 60 * 60 * 24
        elif half_life == "week":
            half_life = 60 * 60 * 24 * 7
        else:
            raise CredoError("Unknown half life value", half_life=half_life)

    return half_life

class SignedValueFile(object):
    """Knows how to store a value in a file with a signature"""
    def __init__(self, location, crypto, extra_info):
        self.crypto = crypto
        self.location = location
        self.extra_info = extra_info

    def retrieve(self, name, question, suggestions):
        """Return the value in our signed file"""
        value = self.recorded_value()
        if not value:
            value = self.ask_for_value(name, question, suggestions)
            fingerprint, signature = self.crypto.create_signature(self.signature_value(value).encode('utf-8'))

            dirname = os.path.dirname(self.location)
            if not os.path.exists(dirname):
                os.makedirs(dirname)

            with open(self.location, "w") as fle:
                fle.write("{0},{1},{2}".format(value, fingerprint, signature))
            return value, True
        return value, False

    def signature_value(self, value):
        """Return string for signing in the signed value file"""
        infos = "|".join(str(thing) for _, thing in sorted(self.extra_info.items()))
        return "{0}|{1}".format(infos, value)

    def recorded_value(self):
        """Read our current signed value"""
        incorrect = False
        value = None

        if os.path.exists(self.location) and os.access(self.location, os.R_OK):
            with open(self.location) as fle:
                contents = fle.read().strip().split("\n")[0]

            if contents.count(',') != 2:
                incorrect = True
            else:
                value, fingerprint, signature = contents.split(',')
                if not self.crypto.is_signature_valid(self.signature_value(value).encode('utf-8'), fingerprint, signature):
                    incorrect = True

        if incorrect:
            log.error("Was something corrupt about the value file\tlocation=%s", self.location)
            return

        return value

    def ask_for_value(self, name, question, suggestions):
        """Get a value from the user"""
        choices = ["Quit"]

        if callable(suggestions):
            suggestions = suggestions()

        suggestion_strings = []
        if suggestions:
            for suggestion in suggestions:
                suggestion_string = str(suggestion)
                suggestion_strings.append(suggestion_string)
                choices.insert(0, suggestion_string)

        choice = ask_for_choice_or_new(question, choices)

        if choice == "Quit":
            raise NoValueEntered()
        else:
            value = choice

        return value

class KeysFile(object):
    """Understands how to load and save to a keys file"""
    def __init__(self, default_keys_type=list, default_keys_type_name="amazon"):
        self.default_keys_type = default_keys_type
        self.default_keys_type_name = default_keys_type_name

    def read_file(self, location):
        """Read in our location as a json file"""
        if not os.path.exists(location):
            raise BadKeyFile("Doesn't exist", location=location)
        if not os.access(location, os.R_OK):
            raise BadKeyFile("Don't have read permissions", location=location)

        if os.stat(location).st_size == 0:
            return {}

        try:
            return json.load(open(location))
        except ValueError as err:
            raise BadKeyFile("Keys file not valid json", location=location, error=err)

    def load(self, location):
        """Load the keys from our keys file"""
        contents = {"keys": self.default_keys_type(), "type": self.default_keys_type_name}
        if os.path.exists(location):
            contents = self.read_file(location)

        if not isinstance(contents.get("keys", []), self.default_keys_type):
            raise BadKeyFile("Keys file keys are not correct", expected_type=self.default_keys_type, got_type=type(contents["keys"]))

        self.contents = contents
        self.typ = self.contents.get("type", self.default_keys_type_name)
        self.keys = self.contents.get("keys", self.default_keys_type())

    def save(self, location, keys, **info):
        """Save the provided keys to this location"""
        dirname = os.path.dirname(location)
        if not os.path.exists(dirname):
            try:
                os.makedirs(dirname)
            except OSError as err:
                raise BadKeyFile("Can't create parent directory", err=err)

        if not os.access(dirname, os.W_OK):
            raise BadKeyFile("Don't have write permissions to parent directory", location=location)

        try:
            key_vals, access_keys = keys.encrypted_values
        except UnicodeDecodeError as err:
            raise BadKeyFile("Can't get encrypted values for the keys file!", err=err, location=location)

        try:
            vals = {"type": keys.type, "keys": key_vals}
            contents = json.dumps(vals, indent=4)
        except ValueError as err:
            raise BadKeyFile("Can't create keys as json", err=err, location=location)

        try:
            with open(location, "w") as fle:
                info_str = ""
                if info:
                    info_str = "\t{0}".format("\t".join("{0}={1}".format(key, val) for key, val in info.items()))
                log.info("Saving keys to %s%s", location, info_str)
                fle.write(contents)
        except OSError as err:
            raise BadKeyFile("Can't write to the keys file", err=err, location=location)

